import unittest
import os
from fontParts.base import normalizers


class TestNormalizers(unittest.TestCase):
    # ----
    # Font
    # ----

    def getFont_layers(self):
        font, _ = self.objectGenerator("font")
        for name in ["A", "B", "C", "D", "E"]:
            font.newLayer(name)
        return font

    # normalizeFileFormatVersion

    def test_normalizeFileFormatVersion_int(self):
        result = normalizers.normalizeFileFormatVersion(3)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 3)

    def test_normalizeFileFormatVersion_float(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeFileFormatVersion(3.0)

    def test_normalizeFileFormatVersion_invalid(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeFileFormatVersion("3")

    # normalizeLayerOrder

    def test_normalizeLayerOrder_valid(self):
        font = self.getFont_layers()
        result = normalizers.normalizeLayerOrder(["A", "B", "C", "D", "E"], font)
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, str)
        self.assertEqual(result, ("A", "B", "C", "D", "E"))

    def test_normalizeLayerOrder_validTuple(self):
        font = self.getFont_layers()
        result = normalizers.normalizeLayerOrder(tuple(["A", "B", "C", "D", "E"]), font)
        self.assertEqual(result, ("A", "B", "C", "D", "E"))

    def test_normalizeLayerOrder_notList(self):
        font = self.getFont_layers()
        with self.assertRaises(TypeError):
            normalizers.normalizeLayerOrder("A B C D E", font)

    def test_normalizeLayerOrder_invalidMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphOrder(["A", "B", "C", "D", 2])

    def test_normalizeLayerOrder_notInFont(self):
        font = self.getFont_layers()
        with self.assertRaises(ValueError):
            normalizers.normalizeLayerOrder(["A", "B", "C", "D", "E", "X"], font)

    def test_normalizeLayerOrder_duplicate(self):
        font = self.getFont_layers()
        with self.assertRaises(ValueError):
            normalizers.normalizeLayerOrder(["A", "B", "C", "C", "D", "E"], font)

    # normalizeDefaultLayerName

    def test_normalizeDefaultLayerName_valid(self):
        font = self.getFont_layers()
        result = normalizers.normalizeDefaultLayerName("B", font)
        self.assertIsInstance(result, str)
        self.assertEqual(result, "B")

    def test_normalizeDefaultLayerName_notValidLayerName(self):
        font = self.getFont_layers()
        with self.assertRaises(TypeError):
            normalizers.normalizeDefaultLayerName(1, font)

    def test_normalizeDefaultLayerName_notInFont(self):
        font = self.getFont_layers()
        with self.assertRaises(ValueError):
            normalizers.normalizeDefaultLayerName("X", font)

    # normalizeGlyphOrder

    def test_normalizeGlyphOrder_valid(self):
        result = normalizers.normalizeGlyphOrder(["A", "B", "C", "D", "E"])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, str)
        self.assertEqual(result, ("A", "B", "C", "D", "E"))

    def test_normalizeGlyphOrder_validTuple(self):
        result = normalizers.normalizeGlyphOrder(tuple(["A", "B", "C", "D", "E"]))
        self.assertEqual(result, ("A", "B", "C", "D", "E"))

    def test_normalizeGlyphOrder_notList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphOrder("A B C D E")

    def test_normalizeGlyphOrder_invalidMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphOrder(["A", "B", "C", "D", 2])

    def test_normalizeGlyphOrder_duplicate(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphOrder(["A", "B", "C", "C", "D", "E"])

    # -------
    # Kerning
    # -------

    # normalizeKerningKey

    def test_normalizeKerningKey_validGlyphs(self):
        result = normalizers.normalizeKerningKey(("A", "B"))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, ("A", "B"))

    def test_normalizeKerningKey_validGroups(self):
        result = normalizers.normalizeKerningKey(("public.kern1.A", "public.kern2.B"))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, ("public.kern1.A", "public.kern2.B"))

    def test_normalizeKerningKey_validList(self):
        result = normalizers.normalizeKerningKey(["A", "B"])
        self.assertEqual(result, ("A", "B"))

    def test_normalizeKerningKey_notTuple(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeKerningKey("A B")

    def test_normalizeKerningKey_notEnoughMembers(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeKerningKey(("A",))

    def test_normalizeKerningKey_tooManyMembers(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeKerningKey(("A", "B", "C"))

    def test_normalizeKerningKey_memberNotString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeKerningKey(("A", 2))
        with self.assertRaises(TypeError):
            normalizers.normalizeKerningKey((1, "B"))

    def test_normalizeKerningKey_memberNotLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeKerningKey(("A", ""))
        with self.assertRaises(ValueError):
            normalizers.normalizeKerningKey(("", "B"))

    def test_normalizeKerningKey_invalidSide1Group(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeKerningKey(("public.kern2.A", "B"))

    def test_normalizeKerningKey_invalidSide2Group(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeKerningKey(("A", "public.kern1.B"))

    # normalizeKerningValue

    def test_normalizeKerningValue_zero(self):
        result = normalizers.normalizeKerningValue(0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 0)

    def test_normalizeKerningValue_positiveInt(self):
        result = normalizers.normalizeKerningValue(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeKerningValue_negativeInt(self):
        result = normalizers.normalizeKerningValue(-1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, -1)

    def test_normalizeKerningValue_positiveFloat(self):
        result = normalizers.normalizeKerningValue(1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeKerningValue_negativeFloat(self):
        result = normalizers.normalizeKerningValue(-1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, -1.0)

    def test_normalizeKerningValue_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeKerningValue("1")

    # ------
    # Groups
    # ------

    # normalizeGroupKey

    def test_normalizeGroupKey_valid(self):
        result = normalizers.normalizeGroupKey("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeGroupKey_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGroupKey(1)

    def test_normalizeGroupKey_notLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGroupKey("")

    # normalizeGroupValue

    def test_normalizeGroupValue_valid(self):
        result = normalizers.normalizeGroupValue(["A", "B", "C"])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, ("A", "B", "C"))

    def test_normalizeGroupValue_validTuple(self):
        result = normalizers.normalizeGroupValue(("A", "B", "C"))
        self.assertEqual(result, ("A", "B", "C"))

    def test_normalizeGroupValue_validEmpty(self):
        result = normalizers.normalizeGroupValue([])
        self.assertEqual(result, tuple())

    def test_normalizeGroupValue_notList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGroupValue("A B C")

    def test_normalizeGroupValue_invalidMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGroupValue(["A", "B", 3])

    # --------
    # Features
    # --------

    # normalizeFeatureText

    def test_normalizeFeatureText_valid(self):
        result = normalizers.normalizeFeatureText("test")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "test")

    def test_normalizeFeatureText_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeFeatureText(123)

    # ---
    # Lib
    # ---

    # normalizeLibKey

    def test_normalizeLibKey_valid(self):
        result = normalizers.normalizeLibKey("test")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "test")

    def test_normalizeLibKey_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLibKey(123)

    def test_normalizeLibKey_emptyString(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeLibKey("")

    # normalizeLibValue

    def test_normalizeLibValue_invalidNone(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLibValue(None)

    def test_normalizeLibValue_validString(self):
        result = normalizers.normalizeLibValue("test")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "test")

    def test_normalizeLibValue_validInt(self):
        result = normalizers.normalizeLibValue(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeLibValue_validFloat(self):
        result = normalizers.normalizeLibValue(1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeLibValue_validTuple(self):
        result = normalizers.normalizeLibValue(("A", "B"))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, ("A", "B"))

    def test_normalizeLibValue_invalidTupleMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLibValue((1, None))

    def test_normalizeLibValue_validList(self):
        result = normalizers.normalizeLibValue(["A", "B"])
        self.assertIsInstance(result, list)
        self.assertEqual(result, ["A", "B"])

    def test_normalizeLibValue_invalidListMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLibValue([1, None])

    def test_normalizeLibValue_validDict(self):
        result = normalizers.normalizeLibValue({"A": 1, "B": 2})
        self.assertIsInstance(result, dict)
        self.assertEqual(result, {"A": 1, "B": 2})

    def test_normalizeLibValue_invalidDictKey(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLibValue({1: 1, "B": 2})

    def test_normalizeLibValue_invalidDictValue(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLibValue({"A": None, "B": 2})

    # -----
    # Layer
    # -----

    # normalizeLayer

    def test_normalizeLayer_valid(self):
        from fontParts.base.layer import BaseLayer

        layer, _ = self.objectGenerator("layer")
        result = normalizers.normalizeLayer(layer)
        self.assertIsInstance(result, BaseLayer)
        self.assertEqual(result, layer)

    def test_normalizeLayer_notLayer(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLayer(123)

    # normalizeLayerName

    def test_normalizeLayerName_valid(self):
        result = normalizers.normalizeLayerName("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeLayerName_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeLayerName(123)

    def test_normalizeLayerName_notLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeLayerName("")

    # -----
    # Glyph
    # -----

    # normalizeGlyph

    def test_normalizeGlyph_valid(self):
        from fontParts.base.glyph import BaseGlyph

        glyph, _ = self.objectGenerator("glyph")
        result = normalizers.normalizeGlyph(glyph)
        self.assertIsInstance(result, BaseGlyph)
        self.assertEqual(result, glyph)

    def test_normalizeGlyph_notGlyph(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyph(123)

    # normalizeGlyphName

    def test_normalizeGlyphName_valid(self):
        result = normalizers.normalizeGlyphName("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeGlyphName_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphName(123)

    def test_normalizeGlyphName_notLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphName("")

    # normalizeGlyphUnicodes

    def test_normalizeGlyphUnicodes_valid(self):
        result = normalizers.normalizeGlyphUnicodes([1, 2, 3])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1, 2, 3))

    def test_normalizeGlyphUnicodes_validTuple(self):
        result = normalizers.normalizeGlyphUnicodes(tuple([1, 2, 3]))
        self.assertEqual(result, (1, 2, 3))

    def test_normalizeGlyphUnicodes_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphUnicodes("xyz")

    def test_normalizeGlyphUnicodes_invalidDuplicateMembers(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphUnicodes([1, 2, 3, 2])

    def test_normalizeGlyphUnicodes_invalidMember(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphUnicodes([1, 2, "xyz"])

    # normalizeGlyphUnicode

    def test_normalizeGlyphUnicode_validInt(self):
        result = normalizers.normalizeGlyphUnicode(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeGlyphUnicode_validHex(self):
        result = normalizers.normalizeGlyphUnicode("0001")
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeGlyphUnicode_validRangeMinimum(self):
        result = normalizers.normalizeGlyphUnicode(0)
        self.assertEqual(result, 0)

    def test_normalizeGlyphUnicode_validRangeMaximum(self):
        result = normalizers.normalizeGlyphUnicode(1114111)
        self.assertEqual(result, 1114111)

    def test_normalizeGlyphUnicode_invalidFloat(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphUnicode(1.0)

    def test_normalizeGlyphUnicode_invalidString(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphUnicode("xyz")

    def test_normalizeGlyphUnicode_invalidRangeMinimum(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphUnicode(-1)

    def test_normalizeGlyphUnicode_invalidRangeMaximum(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphUnicode(1114112)

    # normalizeGlyphBottomMargin

    def test_normalizeGlyphBottomMargin_zero(self):
        result = normalizers.normalizeGlyphBottomMargin(0)
        self.assertEqual(result, 0)

    def test_normalizeGlyphBottomMargin_positiveInt(self):
        result = normalizers.normalizeGlyphBottomMargin(1)
        self.assertEqual(result, 1)

    def test_normalizeGlyphBottomMargin_negativeInt(self):
        result = normalizers.normalizeGlyphBottomMargin(-1)
        self.assertEqual(result, -1)

    def test_normalizeGlyphBottomMargin_positiveFloat(self):
        result = normalizers.normalizeGlyphBottomMargin(1.01)
        self.assertEqual(result, 1.01)

    def test_normalizeGlyphBottomMargin_negativeFloat(self):
        result = normalizers.normalizeGlyphBottomMargin(-1.01)
        self.assertEqual(result, -1.01)

    def test_normalizeGlyphBottomMargin_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphBottomMargin("1")

    # normalizeGlyphLeftMargin

    def test_normalizeGlyphLeftMargin_zero(self):
        result = normalizers.normalizeGlyphLeftMargin(0)
        self.assertEqual(result, 0)

    def test_normalizeGlyphLeftMargin_positiveInt(self):
        result = normalizers.normalizeGlyphLeftMargin(1)
        self.assertEqual(result, 1)

    def test_normalizeGlyphLeftMargin_negativeInt(self):
        result = normalizers.normalizeGlyphLeftMargin(-1)
        self.assertEqual(result, -1)

    def test_normalizeGlyphLeftMargin_positiveFloat(self):
        result = normalizers.normalizeGlyphLeftMargin(1.01)
        self.assertEqual(result, 1.01)

    def test_normalizeGlyphLeftMargin_negativeFloat(self):
        result = normalizers.normalizeGlyphLeftMargin(-1.01)
        self.assertEqual(result, -1.01)

    def test_normalizeGlyphLeftMargin_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphLeftMargin("1")

    # normalizeGlyphRightMargin

    def test_normalizeGlyphRightMargin_zero(self):
        result = normalizers.normalizeGlyphRightMargin(0)
        self.assertEqual(result, 0)

    def test_normalizeGlyphRightMargin_positiveInt(self):
        result = normalizers.normalizeGlyphRightMargin(1)
        self.assertEqual(result, 1)

    def test_normalizeGlyphRightMargin_negativeInt(self):
        result = normalizers.normalizeGlyphRightMargin(-1)
        self.assertEqual(result, -1)

    def test_normalizeGlyphRightMargin_positiveFloat(self):
        result = normalizers.normalizeGlyphRightMargin(1.01)
        self.assertEqual(result, 1.01)

    def test_normalizeGlyphRightMargin_negativeFloat(self):
        result = normalizers.normalizeGlyphRightMargin(-1.01)
        self.assertEqual(result, -1.01)

    def test_normalizeGlyphRightMargin_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphRightMargin("1")

    # normalizeGlyphHeight

    def test_normalizeGlyphHeight_zero(self):
        result = normalizers.normalizeGlyphHeight(0)
        self.assertEqual(result, 0)

    def test_normalizeGlyphHeight_positiveInt(self):
        result = normalizers.normalizeGlyphHeight(1)
        self.assertEqual(result, 1)

    def test_normalizeGlyphHeight_negativeInt(self):
        result = normalizers.normalizeGlyphHeight(-1)
        self.assertEqual(result, -1)

    def test_normalizeGlyphHeight_positiveFloat(self):
        result = normalizers.normalizeGlyphHeight(1.01)
        self.assertEqual(result, 1.01)

    def test_normalizeGlyphHeight_negativeFloat(self):
        result = normalizers.normalizeGlyphHeight(-1.01)
        self.assertEqual(result, -1.01)

    def test_normalizeGlyphHeight_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphHeight("1")

    # normalizeGlyphTopMargin

    def test_normalizeGlyphTopMargin_zero(self):
        result = normalizers.normalizeGlyphTopMargin(0)
        self.assertEqual(result, 0)

    def test_normalizeGlyphTopMargin_positiveInt(self):
        result = normalizers.normalizeGlyphTopMargin(1)
        self.assertEqual(result, 1)

    def test_normalizeGlyphTopMargin_negativeInt(self):
        result = normalizers.normalizeGlyphTopMargin(-1)
        self.assertEqual(result, -1)

    def test_normalizeGlyphTopMargin_positiveFloat(self):
        result = normalizers.normalizeGlyphTopMargin(1.01)
        self.assertEqual(result, 1.01)

    def test_normalizeGlyphTopMargin_negativeFloat(self):
        result = normalizers.normalizeGlyphTopMargin(-1.01)
        self.assertEqual(result, -1.01)

    def test_normalizeGlyphTopMargin_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphTopMargin("1")

    # normalizeGlyphFormatVersion

    def test_normalizeGlyphFormatVersion_int1(self):
        result = normalizers.normalizeGlyphFormatVersion(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeGlyphFormatVersion_int2(self):
        result = normalizers.normalizeGlyphFormatVersion(2)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 2)

    def test_normalizeGlyphFormatVersion_int3(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphFormatVersion(3)

    def test_normalizeGlyphFormatVersion_float1(self):
        result = normalizers.normalizeGlyphFormatVersion(1.0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeGlyphFormatVersion_float2(self):
        result = normalizers.normalizeGlyphFormatVersion(2.0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 2)

    def test_normalizeGlyphFormatVersion_float3(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGlyphFormatVersion(3.0)

    def test_normalizeGlyphFormatVersion_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphFormatVersion("1")

    # -------
    # Contour
    # -------

    # normalizeContour

    def test_normalizeContour_valid(self):
        from fontParts.base.contour import BaseContour

        contour, _ = self.objectGenerator("contour")
        result = normalizers.normalizeContour(contour)
        self.assertIsInstance(result, BaseContour)
        self.assertEqual(result, contour)

    def test_normalizeContour_notContour(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeContour(123)

    # -----
    # Point
    # -----

    # normalizePoint

    def test_normalizePoint_valid(self):
        from fontParts.base.point import BasePoint

        point, _ = self.objectGenerator("point")
        result = normalizers.normalizePoint(point)
        self.assertIsInstance(result, BasePoint)
        self.assertEqual(result, point)

    def test_normalizePoint_notPoint(self):
        with self.assertRaises(TypeError):
            normalizers.normalizePoint(123)

    # normalizePointType

    def test_normalizePointType_move(self):
        result = normalizers.normalizePointType("move")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "move")

    def test_normalizePointType_Move(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("Move")

    def test_normalizePointType_MOVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("MOVE")

    def test_normalizePointType_line(self):
        result = normalizers.normalizePointType("line")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "line")

    def test_normalizePointType_Line(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("Line")

    def test_normalizePointType_LINE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("LINE")

    def test_normalizePointType_offcurve(self):
        result = normalizers.normalizePointType("offcurve")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "offcurve")

    def test_normalizePointType_OffCurve(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("OffCurve")

    def test_normalizePointType_OFFCURVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("OFFCURVE")

    def test_normalizePointType_curve(self):
        result = normalizers.normalizePointType("curve")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "curve")

    def test_normalizePointType_Curve(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("Curve")

    def test_normalizePointType_CURVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("CURVE")

    def test_normalizePointType_qcurve(self):
        result = normalizers.normalizePointType("qcurve")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "qcurve")

    def test_normalizePointType_QOffCurve(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("QCurve")

    def test_normalizePointType_QOFFCURVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("QCURVE")

    def test_normalizePointType_unknown(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointType("unknonwn")

    def test_normalizePointType_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizePointType(123)

    # normalizePointName

    def test_normalizePointName_valid(self):
        result = normalizers.normalizePointName("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizePointName_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizePointName(123)

    def test_normalizePointName_notLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizePointName("")

    # -------
    # Segment
    # -------

    # normalizeSegment

    def test_normalizeSegment_valid(self):
        from fontParts.base.segment import BaseSegment

        segment, _ = self.objectGenerator("segment")
        result = normalizers.normalizeSegment(segment)
        self.assertIsInstance(result, BaseSegment)
        self.assertEqual(result, segment)

    def test_normalizePoint_notContour(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeSegment(123)

    # normalizeSegmentType

    def test_normalizeSegmentType_move(self):
        result = normalizers.normalizeSegmentType("move")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "move")

    def test_normalizeSegmentType_Move(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("Move")

    def test_normalizeSegmentType_MOVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("MOVE")

    def test_normalizeSegmentType_line(self):
        result = normalizers.normalizeSegmentType("line")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "line")

    def test_normalizeSegmentType_Line(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("Line")

    def test_normalizeSegmentType_LINE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("LINE")

    def test_normalizeSegmentType_curve(self):
        result = normalizers.normalizeSegmentType("curve")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "curve")

    def test_normalizeSegmentType_OffCurve(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("Curve")

    def test_normalizeSegmentType_OFFCURVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("CURVE")

    def test_normalizeSegmentType_qcurve(self):
        result = normalizers.normalizeSegmentType("qcurve")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "qcurve")

    def test_normalizeSegmentType_QOffCurve(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("QCurve")

    def test_normalizeSegmentType_QOFFCURVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("QCURVE")

    def test_normalizeSegmentType_unknown(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeSegmentType("offcurve")

    def test_normalizeSegmentType_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeSegmentType(123)

    # ------
    # BPoint
    # ------

    # normalizeBPoint

    def test_normalizeBPoint_valid(self):
        from fontParts.base.bPoint import BaseBPoint

        bPoint, _ = self.objectGenerator("bPoint")
        result = normalizers.normalizeBPoint(bPoint)
        self.assertIsInstance(result, BaseBPoint)
        self.assertEqual(result, bPoint)

    def test_normalizeBPoint_notBPoint(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeBPoint(123)

    # normalizeBPointType

    def test_normalizeBPointType_corner(self):
        result = normalizers.normalizeBPointType("corner")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "corner")

    def test_normalizeBPointType_Corner(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBPointType("Corner")

    def test_normalizeBPointType_CORNER(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBPointType("CORNER")

    def test_normalizeBPointType_curve(self):
        result = normalizers.normalizeBPointType("curve")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "curve")

    def test_normalizeBPointType_OffCurve(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBPointType("Curve")

    def test_normalizeBPointType_OFFCURVE(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBPointType("CURVE")

    def test_normalizeBPointType_unknown(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBPointType("offcurve")

    def test_normalizeBPointType_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeBPointType(123)

    # ---------
    # Component
    # ---------

    # normalizeComponent

    def test_normalizeComponent_valid(self):
        from fontParts.base.component import BaseComponent

        component, _ = self.objectGenerator("component")
        result = normalizers.normalizeComponent(component)
        self.assertIsInstance(result, BaseComponent)
        self.assertEqual(result, component)

    def test_normalizeComponent_notComponent(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeComponent(123)

    # normalizeComponentScale

    def test_normalizeComponentScale_tupleZero(self):
        result = normalizers.normalizeComponentScale((0, 0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (0, 0))

    def test_normalizeComponentScale_tuplePositiveInt(self):
        result = normalizers.normalizeComponentScale((2, 2))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeComponentScale_tupleNegativeInt(self):
        result = normalizers.normalizeComponentScale((-2, -2))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeComponentScale_tuplePositiveFloat(self):
        result = normalizers.normalizeComponentScale((2.0, 2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeComponentScale_tupleNegativeFloat(self):
        result = normalizers.normalizeComponentScale((-2.0, -2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeComponentScale_listZero(self):
        result = normalizers.normalizeComponentScale([0, 0])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (0, 0))

    def test_normalizeComponentScale_listPositiveInt(self):
        result = normalizers.normalizeComponentScale([2, 2])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeComponentScale_listNegativeInt(self):
        result = normalizers.normalizeComponentScale([-2, -2])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeComponentScale_listPositiveFloat(self):
        result = normalizers.normalizeComponentScale([2.0, 2.0])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeComponentScale_listNegativeFloat(self):
        result = normalizers.normalizeComponentScale([-2.0, -2.0])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeComponentScale_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeComponentScale("2, 2")

    def test_normalizeComponentScale_numberNotNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeComponentScale((2, "2"))

    def test_normalizeComponentScale_notNumberNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeComponentScale(("2", 2))

    def test_normalizeComponentScale_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeComponentScale((2,))

    def test_normalizeComponentScale_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeComponentScale((2, 2, 2))

    # ------
    # Anchor
    # ------

    # normalizeAnchor

    def test_normalizeAnchor_valid(self):
        from fontParts.base.anchor import BaseAnchor

        anchor, _ = self.objectGenerator("anchor")
        result = normalizers.normalizeAnchor(anchor)
        self.assertIsInstance(result, BaseAnchor)
        self.assertEqual(result, anchor)

    def test_normalizeAnchor_notAnchor(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeAnchor(123)

    # normalizeAnchorName

    def test_normalizeAnchorName_valid(self):
        result = normalizers.normalizeAnchorName("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeAnchorName_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeAnchorName(123)

    def test_normalizeAnchorName_notLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeAnchorName("")

    # ---------
    # Guideline
    # ---------

    # normalizeGuideline

    def test_normalizeGuideline_valid(self):
        from fontParts.base.guideline import BaseGuideline

        guideline, _ = self.objectGenerator("guideline")
        result = normalizers.normalizeGuideline(guideline)
        self.assertIsInstance(result, BaseGuideline)
        self.assertEqual(result, guideline)

    def test_normalizeGuideline_notGuideline(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGuideline(123)

    # normalizeGuidelineName

    def test_normalizeGuidelineName_valid(self):
        result = normalizers.normalizeGuidelineName("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeGuidelineName_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGuidelineName(123)

    def test_normalizeGuidelineName_notLongEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeGuidelineName("")

    # -------
    # Generic
    # -------

    # normalizeBoolean

    def test_normalizeBoolean_true(self):
        result = normalizers.normalizeBoolean(True)
        self.assertIsInstance(result, bool)
        self.assertEqual(result, True)

    def test_normalizeBoolean_false(self):
        result = normalizers.normalizeBoolean(False)
        self.assertIsInstance(result, bool)
        self.assertEqual(result, False)

    def test_normalizeBoolean_1(self):
        result = normalizers.normalizeBoolean(1)
        self.assertIsInstance(result, bool)
        self.assertEqual(result, True)

    def test_normalizeBoolean_0(self):
        result = normalizers.normalizeBoolean(0)
        self.assertIsInstance(result, bool)
        self.assertEqual(result, False)

    def test_normalizeBoolean_10(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBoolean(10)

    def test_normalizeBoolean_string(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBoolean("True")

    # normalizeIndex

    def test_normalizeIndex_zero(self):
        result = normalizers.normalizeIndex(0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 0)

    def test_normalizeIndex_positiveInt(self):
        result = normalizers.normalizeIndex(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeIndex_negativeInt(self):
        result = normalizers.normalizeIndex(-1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, -1)

    def test_normalizeIndex_notInt(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeIndex(1.0)

    # normalizeIdentifier

    def test_normalizeIdentifier_none(self):
        result = normalizers.normalizeIdentifier(None)
        self.assertEqual(result, None)

    def test_normalizeIdentifier_stringMinimumLength(self):
        result = normalizers.normalizeIdentifier("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeIdentifier_stringMaximumLength(self):
        result = normalizers.normalizeIdentifier("A" * 100)
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A" * 100)

    def test_normalizeIdentifier_stringMinimumCharacter(self):
        result = normalizers.normalizeIdentifier(chr(0x20))
        self.assertIsInstance(result, str)
        self.assertEqual(result, chr(0x20))

    def test_normalizeIdentifier_stringMaximumCharacter(self):
        result = normalizers.normalizeIdentifier(chr(0x7E))
        self.assertIsInstance(result, str)
        self.assertEqual(result, chr(0x7E))

    def test_normalizeIdentifier_stringTooShort(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeIdentifier("")

    def test_normalizeIdentifier_stringTooLong(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeIdentifier("A" * 101)

    def test_normalizeIdentifier_stringBeforeMinimumCharacter(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeIdentifier(chr(0x20 - 1))

    def test_normalizeIdentifier_stringAfterMaximumCharacter(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeIdentifier(chr(0x7E + 1))

    def test_normalizeIdentifier_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeIdentifier(1)

    # normalizeX

    def test_normalizeX_zero(self):
        result = normalizers.normalizeX(0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 0)

    def test_normalizeX_positiveInt(self):
        result = normalizers.normalizeX(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeX_negativeInt(self):
        result = normalizers.normalizeX(-1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, -1)

    def test_normalizeX_positiveFloat(self):
        result = normalizers.normalizeX(1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeX_negativeFloat(self):
        result = normalizers.normalizeX(-1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, -1.0)

    def test_normalizeX_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeX("1")

    # normalizeY

    def test_normalizeY_zero(self):
        result = normalizers.normalizeY(0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 0)

    def test_normalizeY_positiveInt(self):
        result = normalizers.normalizeY(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeY_negativeInt(self):
        result = normalizers.normalizeY(-1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, -1)

    def test_normalizeY_positiveFloat(self):
        result = normalizers.normalizeY(1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeY_negativeFloat(self):
        result = normalizers.normalizeY(-1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, -1.0)

    def test_normalizeY_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeY("1")

    # normalizeCoordinateTuple

    def test_normalizeCoordinateTuple_list(self):
        result = normalizers.normalizeCoordinateTuple([1, 1])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1, 1))

    def test_normalizeCoordinateTuple_zero(self):
        result = normalizers.normalizeCoordinateTuple((0, 0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0))

    def test_normalizeCoordinateTuple_positiveInt(self):
        result = normalizers.normalizeCoordinateTuple((1, 1))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1, 1))

    def test_normalizeCoordinateTuple_negativeInt(self):
        result = normalizers.normalizeCoordinateTuple((-1, -1))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-1, -1))

    def test_normalizeCoordinateTuple_positiveFloat(self):
        result = normalizers.normalizeCoordinateTuple((1.0, 1.0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 1.0))

    def test_normalizeCoordinateTuple_negativeFloat(self):
        result = normalizers.normalizeCoordinateTuple((-1.0, -1.0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-1.0, -1.0))

    def test_normalizeCoordinateTuple_numberNotNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeCoordinateTuple((1, "1"))

    def test_normalizeCoordinateTuple_notNumberNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeCoordinateTuple(("1", 1))

    def test_normalizeCoordinateTuple_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeCoordinateTuple((1,))

    def test_normalizeCoordinateTuple_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeCoordinateTuple((1, 1, 1))

    # normalizeBoundingBox

    def test_normalizeBoundingBox_tuple(self):
        result = normalizers.normalizeBoundingBox((1, 2, 3, 4))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0))

    def test_normalizeBoundingBox_list(self):
        result = normalizers.normalizeBoundingBox([1, 2, 3, 4])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0))

    def test_normalizeBoundingBox_int(self):
        result = normalizers.normalizeBoundingBox((1, 2, 3, 4))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0))

    def test_normalizeBoundingBox_float(self):
        result = normalizers.normalizeBoundingBox((1.0, 2.0, 3.0, 4.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0))

    def test_normalizeBoundingBox_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBoundingBox((1, 2, 3))

    def test_normalizeBoundingBox_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBoundingBox((1, 2, 3, 4, 5))

    def test_normalizeBoundingBox_notListOrTuple(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeBoundingBox("1 2 3 4")

    def test_normalizeBoundingBox_invalidMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeBoundingBox((1, 2, "3", 4))

    def test_normalizeBoundingBox_xMinSameAsXMax(self):
        result = normalizers.normalizeBoundingBox((1, 2, 1, 4))
        self.assertEqual(result, (1.0, 2.0, 1.0, 4.0))

    def test_normalizeBoundingBox_yMinSameAsYMax(self):
        result = normalizers.normalizeBoundingBox((1, 2, 3, 2))
        self.assertEqual(result, (1.0, 2.0, 3.0, 2.0))

    def test_normalizeBoundingBox_xMaxLessThanXMin(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBoundingBox((1, 2, 0, 4))

    def test_normalizeBoundingBox_yMaxLessThanYMin(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeBoundingBox((1, 2, 3, 1))

    # normalizeArea

    def test_normalizeArea_zero(self):
        result = normalizers.normalizeArea(0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 0)

    def test_normalizeArea_positiveInt(self):
        result = normalizers.normalizeArea(1)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeArea_positiveFloat(self):
        result = normalizers.normalizeArea(1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeArea_negativeInt(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeArea(-1)

    def test_normalizeArea_negativeFloat(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeArea(-1.0)

    def test_normalizeArea_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeArea("1")

    # normalizeColor

    def test_normalizeColor_color(self):
        from fontParts.base.color import Color

        result = normalizers.normalizeColor(Color((0, 0, 0, 0)))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0, 0, 0))

    def test_normalizeColor_tuple(self):
        result = normalizers.normalizeColor((0, 0, 0, 0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0, 0, 0))

    def test_normalizeColor_list(self):
        result = normalizers.normalizeColor([0, 0, 0, 0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0, 0, 0))

    def test_normalizeColor_ints(self):
        result = normalizers.normalizeColor((1, 1, 1, 1))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 1.0, 1.0, 1.0))

    def test_normalizeColor_floats(self):
        result = normalizers.normalizeColor((1.0, 1.0, 1.0, 1.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 1.0, 1.0, 1.0))

    def test_normalizeColor_tooSmall(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((-1, 1, 1, 1))
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, -1, 1, 1))
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 1, -1, 1))
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 1, 1, -1))

    def test_normalizeColor_tooLarge(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((2, 1, 1, 1))
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 2, 1, 1))
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 1, 2, 1))
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 1, 1, 2))

    def test_normalizeColor_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 1, 1))

    def test_normalizeColor_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeColor((1, 1, 1, 1, 1))

    def test_normalizeColor_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeColor("1,1,1,1")

    def test_normalizeColor_invalidMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeColor((1, "1", 1, 1))

    # normalizeGlyphNote

    def test_normalizeGlyphNote_string(self):
        result = normalizers.normalizeGlyphNote("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeGlyphNote_emptyString(self):
        result = normalizers.normalizeGlyphNote("")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "")

    def test_normalizeGlyphNote_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeGlyphNote(123)

    # normalizeFilePath
    def test_normalizeFilePath_pathlib_fillPath(self):
        from pathlib import Path

        path = Path(os.getcwd(), "Test.ufo")
        self.assertIsInstance(path, Path)
        result = normalizers.normalizeFilePath(path)
        self.assertIsInstance(result, str)
        self.assertEqual(result, os.path.join(os.getcwd(), "Test.ufo"))

    def test_normalizeFilePath_pathlib_simplePath(self):
        from pathlib import Path

        path = Path("Test.ufo")
        self.assertIsInstance(path, Path)
        result = normalizers.normalizeFilePath(path)
        self.assertIsInstance(result, str)
        self.assertEqual(result, "Test.ufo")

    def test_normalizeFilePath_string(self):
        result = normalizers.normalizeFilePath("A")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "A")

    def test_normalizeFilePath_emptyString(self):
        result = normalizers.normalizeFilePath("")
        self.assertIsInstance(result, str)
        self.assertEqual(result, "")

    def test_normalizeFilePath_notString(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeFilePath(123)

    # normalizeInterpolationFactor

    def test_normalizeInterpolationFactor_zero(self):
        result = normalizers.normalizeInterpolationFactor(0)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0))

    def test_normalizeInterpolationFactor_positiveInt(self):
        result = normalizers.normalizeInterpolationFactor(1)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 1.0))

    def test_normalizeInterpolationFactor_negativeInt(self):
        result = normalizers.normalizeInterpolationFactor(-1)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-1.0, -1.0))

    def test_normalizeInterpolationFactor_positiveFloat(self):
        result = normalizers.normalizeInterpolationFactor(1.0)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 1.0))

    def test_normalizeInterpolationFactor_negativeFloat(self):
        result = normalizers.normalizeInterpolationFactor(-1.0)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-1.0, -1.0))

    def test_normalizeInterpolationFactor_tupleZero(self):
        result = normalizers.normalizeInterpolationFactor((0, 0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (0, 0))

    def test_normalizeInterpolationFactor_tuplePositiveInt(self):
        result = normalizers.normalizeInterpolationFactor((2, 2))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeInterpolationFactor_tupleNegativeInt(self):
        result = normalizers.normalizeInterpolationFactor((-2, -2))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeInterpolationFactor_tuplePositiveFloat(self):
        result = normalizers.normalizeInterpolationFactor((2.0, 2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeInterpolationFactor_tupleNegativeFloat(self):
        result = normalizers.normalizeInterpolationFactor((-2.0, -2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeInterpolationFactor_listZero(self):
        result = normalizers.normalizeInterpolationFactor([0, 0])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (0, 0))

    def test_normalizeInterpolationFactor_listPositiveInt(self):
        result = normalizers.normalizeInterpolationFactor([2, 2])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeInterpolationFactor_listNegativeInt(self):
        result = normalizers.normalizeInterpolationFactor([-2, -2])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeInterpolationFactor_listPositiveFloat(self):
        result = normalizers.normalizeInterpolationFactor([2.0, 2.0])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeInterpolationFactor_listNegativeFloat(self):
        result = normalizers.normalizeInterpolationFactor([-2.0, -2.0])
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeInterpolationFactor_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeInterpolationFactor("2, 2")

    def test_normalizeInterpolationFactor_numberNotNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeInterpolationFactor((2, "2"))

    def test_normalizeInterpolationFactor_notNumberNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeInterpolationFactor(("2", 2))

    def test_normalizeInterpolationFactor_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeInterpolationFactor((2,))

    def test_normalizeInterpolationFactor_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeInterpolationFactor((2, 2, 2))

    # normalizeRotationAngle

    def test_normalizeRotationAngle_zero(self):
        result = normalizers.normalizeRotationAngle(0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 0)

    def test_normalizeRotationAngle_positiveInt(self):
        result = normalizers.normalizeRotationAngle(1)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeRotationAngle_negativeInt(self):
        result = normalizers.normalizeRotationAngle(-1)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 359.0)

    def test_normalizeRotationAngle_positiveFloat(self):
        result = normalizers.normalizeRotationAngle(1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 1.0)

    def test_normalizeRotationAngle_negativeFloat(self):
        result = normalizers.normalizeRotationAngle(-1.0)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 359.0)

    def test_normalizeRotationAngle_maximum(self):
        result = normalizers.normalizeRotationAngle(360)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 360.0)

    def test_normalizeRotationAngle_minimum(self):
        result = normalizers.normalizeRotationAngle(-360)
        self.assertIsInstance(result, float)
        self.assertEqual(result, 0)

    def test_normalizeRotationAngle_moreThanMaximum(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeRotationAngle(361)

    def test_normalizeRotationAngle_lessThanMaximum(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeRotationAngle(-361)

    def test_normalizeRotationAngle_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeRotationAngle("1")

    # ---------------
    # Transformations
    # ---------------

    # normalizeTransformationMatrix

    def test_normalizeTransformationMatrix_tuple(self):
        result = normalizers.normalizeTransformationMatrix((1, 2, 3, 4, 5, 6))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0, 5.0, 6.0))

    def test_normalizeTransformationMatrix_list(self):
        result = normalizers.normalizeTransformationMatrix([1, 2, 3, 4, 5, 6])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0, 5.0, 6.0))

    def test_normalizeTransformationMatrix_positiveInts(self):
        result = normalizers.normalizeTransformationMatrix((1, 2, 3, 4, 5, 6))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0, 5.0, 6.0))

    def test_normalizeTransformationMatrix_positiveFloats(self):
        result = normalizers.normalizeTransformationMatrix(
            (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)
        )
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 2.0, 3.0, 4.0, 5.0, 6.0))

    def test_normalizeTransformationMatrix_negativeInts(self):
        result = normalizers.normalizeTransformationMatrix((-1, -2, -3, -4, -5, -6))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-1.0, -2.0, -3.0, -4.0, -5.0, -6.0))

    def test_normalizeTransformationMatrix_negativeFloats(self):
        result = normalizers.normalizeTransformationMatrix(
            (-1.0, -2.0, -3.0, -4.0, -5.0, -6.0)
        )
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-1.0, -2.0, -3.0, -4.0, -5.0, -6.0))

    def test_normalizeTransformationMatrix_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationMatrix((1, 2, 3, 4, 5))

    def test_normalizeTransformationMatrix_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationMatrix((1, 2, 3, 4, 5, 6, 7))

    def test_normalizeTransformationMatrix_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationMatrix("1 2 3 4 5 6")

    def test_normalizeTransformationMatrix_invalidMember(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationMatrix((1, 2, 3, "4", 5, 6))

    # normalizeTransformationOffset

    def test_normalizeTransformationOffset_tupleZero(self):
        result = normalizers.normalizeTransformationOffset((0, 0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0))

    def test_normalizeTransformationOffset_tuplePositiveInt(self):
        result = normalizers.normalizeTransformationOffset((2, 2))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationOffset_tupleNegativeInt(self):
        result = normalizers.normalizeTransformationOffset((-2, -2))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationOffset_tuplePositiveFloat(self):
        result = normalizers.normalizeTransformationOffset((2.0, 2.0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationOffset_tupleNegativeFloat(self):
        result = normalizers.normalizeTransformationOffset((-2.0, -2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationOffset_listZero(self):
        result = normalizers.normalizeTransformationOffset([0, 0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0))

    def test_normalizeTransformationOffset_listPositiveInt(self):
        result = normalizers.normalizeTransformationOffset([2, 2])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationOffset_listNegativeInt(self):
        result = normalizers.normalizeTransformationOffset([-2, -2])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationOffset_listPositiveFloat(self):
        result = normalizers.normalizeTransformationOffset([2.0, 2.0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationOffset_listNegativeFloat(self):
        result = normalizers.normalizeTransformationOffset([-2.0, -2.0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationOffset_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationOffset("2, 2")

    def test_normalizeTransformationOffset_numberNotNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationOffset((2, "2"))

    def test_normalizeTransformationOffset_notNumberNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationOffset(("2", 2))

    def test_normalizeTransformationOffset_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationOffset((2,))

    def test_normalizeTransformationOffset_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationOffset((2, 2, 2))

    # normalizeTransformationSkewAngle

    def test_normalizeTransformationSkewAngle_int(self):
        result = normalizers.normalizeTransformationSkewAngle(1)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 0))

    def test_normalizeTransformationSkewAngle_float(self):
        result = normalizers.normalizeTransformationSkewAngle(1.0)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 0))

    def test_normalizeTransformationSkewAngle_list(self):
        result = normalizers.normalizeTransformationSkewAngle([1, 2])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0))

    def test_normalizeTransformationSkewAngle_tuple(self):
        result = normalizers.normalizeTransformationSkewAngle((1, 2))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0))

    def test_normalizeTransformationSkewAngle_zero(self):
        result = normalizers.normalizeTransformationSkewAngle((0, 0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (0, 0))

    def test_normalizeTransformationSkewAngle_positiveInts(self):
        result = normalizers.normalizeTransformationSkewAngle((1, 2))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 2.0))

    def test_normalizeTransformationSkewAngle_negativeInts(self):
        result = normalizers.normalizeTransformationSkewAngle((-1, -2))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (359.0, 358.0))

    def test_normalizeTransformationSkewAngle_positiveFloats(self):
        result = normalizers.normalizeTransformationSkewAngle((1.0, 2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (1.0, 2.0))

    def test_normalizeTransformationSkewAngle_negativeFloats(self):
        result = normalizers.normalizeTransformationSkewAngle((-1.0, -2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (359.0, 358.0))

    def test_normalizeTransformationSkewAngle_tooLow(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationSkewAngle((-361, -361))

    def test_normalizeTransformationSkewAngle_tooHigh(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationSkewAngle((361, 361))

    def test_normalizeTransformationSkewAngle_numberNotNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationSkewAngle((1, "2"))

    def test_normalizeTransformationSkewAngle_notNumberNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationSkewAngle(("1", 1))

    def test_normalizeTransformationSkewAngle_tooFew(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationSkewAngle(tuple())

    def test_normalizeTransformationSkewAngle_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationSkewAngle((1, 2, 3))

    def test_normalizeTransformationSkewAngle_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationSkewAngle("1")

    # normalizeTransformationScale

    def test_normalizeTransformationScale_int(self):
        result = normalizers.normalizeTransformationScale(1)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 1.0))

    def test_normalizeTransformationScale_float(self):
        result = normalizers.normalizeTransformationScale(1.0)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 1.0))

    def test_normalizeTransformationScale_list(self):
        result = normalizers.normalizeTransformationScale([1, 2])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0))

    def test_normalizeTransformationScale_tuple(self):
        result = normalizers.normalizeTransformationScale((1, 2))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (1.0, 2.0))

    def test_normalizeTransformationScale_tupleZero(self):
        result = normalizers.normalizeTransformationScale((0, 0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0))

    def test_normalizeTransformationScale_tuplePositiveInt(self):
        result = normalizers.normalizeTransformationScale((2, 2))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationScale_tupleNegativeInt(self):
        result = normalizers.normalizeTransformationScale((-2, -2))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationScale_tuplePositiveFloat(self):
        result = normalizers.normalizeTransformationScale((2.0, 2.0))
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationScale_tupleNegativeFloat(self):
        result = normalizers.normalizeTransformationScale((-2.0, -2.0))
        self.assertIsInstance(result, tuple)
        for i in result:
            self.assertIsInstance(i, float)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationScale_listZero(self):
        result = normalizers.normalizeTransformationScale([0, 0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (0, 0))

    def test_normalizeTransformationScale_listPositiveInt(self):
        result = normalizers.normalizeTransformationScale([2, 2])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationScale_listNegativeInt(self):
        result = normalizers.normalizeTransformationScale([-2, -2])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationScale_listPositiveFloat(self):
        result = normalizers.normalizeTransformationScale([2.0, 2.0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (2.0, 2.0))

    def test_normalizeTransformationScale_listNegativeFloat(self):
        result = normalizers.normalizeTransformationScale([-2.0, -2.0])
        self.assertIsInstance(result, tuple)
        self.assertEqual(result, (-2.0, -2.0))

    def test_normalizeTransformationScale_notTupleOrList(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationScale("2, 2")

    def test_normalizeTransformationScale_numberNotNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationScale((2, "2"))

    def test_normalizeTransformationScale_notNumberNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeTransformationScale(("2", 2))

    def test_normalizeTransformationScale_notEnough(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationScale((2,))

    def test_normalizeTransformationScale_tooMany(self):
        with self.assertRaises(ValueError):
            normalizers.normalizeTransformationScale((2, 2, 2))

    # normalizeVisualRounding

    def test_normalizeVisualRounding_int(self):
        result = normalizers.normalizeVisualRounding(1)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeVisualRounding_float(self):
        result = normalizers.normalizeVisualRounding(1.0)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_normalizeVisualRounding_half(self):
        result = normalizers.normalizeVisualRounding(1.5)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 2)

    def test_normalizeVisualRounding_half_even(self):
        result = normalizers.normalizeVisualRounding(2.5)
        self.assertIsInstance(result, int)
        self.assertEqual(result, 3)

    def test_normalizeVisualRounding_notNumber(self):
        with self.assertRaises(TypeError):
            normalizers.normalizeVisualRounding("1")



from fontParts.base.errors import FontPartsError
from fontParts.fontshell.font import RFont
from fontParts.fontshell.info import RInfo
from fontParts.fontshell.groups import RGroups
from fontParts.fontshell.kerning import RKerning
from fontParts.fontshell.features import RFeatures
from fontParts.fontshell.lib import RLib
from fontParts.fontshell.layer import RLayer
from fontParts.fontshell.glyph import RGlyph
from fontParts.fontshell.contour import RContour
from fontParts.fontshell.point import RPoint
from fontParts.fontshell.segment import RSegment
from fontParts.fontshell.bPoint import RBPoint
from fontParts.fontshell.component import RComponent
from fontParts.fontshell.anchor import RAnchor
from fontParts.fontshell.guideline import RGuideline
from fontParts.fontshell.image import RImage



from __future__ import annotations
from typing import TYPE_CHECKING, cast, Any, TypeVar
from collections.abc import Callable, Iterator

from fontParts.base.errors import FontPartsError
from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    IdentifierMixin,
    dynamicProperty,
    reference,
)
from fontParts.base import normalizers
from fontParts.base.bounds import Bounds
from fontParts.base.compatibility import ContourCompatibilityReporter
from fontParts.base.deprecated import DeprecatedContour, RemovedContour
from fontParts.base.annotations import (
    AffineTransformationLike,
    BoundingBox,
    CoordinateLike,
    CollectionType,
    PenType,
    PointPenType,
)

if TYPE_CHECKING:
    from fontParts.base.point import BasePoint
    from fontParts.base.bPoint import BaseBPoint
    from fontParts.base.segment import BaseSegment
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.layer import BaseLayer
    from fontParts.base.font import BaseFont

BaseContourType = TypeVar("BaseContourType", bound="BaseContour")
PointCollectionType = CollectionType[CoordinateLike]


class BaseContour(
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    IdentifierMixin,
    DeprecatedContour,
    RemovedContour,
):
    """Represent the basis for a contour object.

    :cvar segmentClass: A class representing contour segments. This will
        usually be a :class:`BaseSegment` subclass.
    :cvar bPointClass: A class representing contour bPoints. This will
        usually be a :class:`BaseBPoint` subclass.

    """

    segmentClass: type[BaseSegment] | None = None
    bPointClass: type[BaseBPoint] | None = None

    def _reprContents(self) -> list[str]:
        contents = []
        if self.identifier is not None:
            contents.append(f"identifier='{self.identifier!r}'")
        if self.glyph is not None:
            contents.append("in glyph")
            contents += self.glyph._reprContents()
        return contents

    def copyData(self, source: BaseContourType) -> None:
        """Copy data from another contour instance.

        This will copy the contents of the following attributes from `source`
        into the current contour instance:

        - :attr:`BaseContour.points`
        - :attr:`BaseContour.bPoints`

        :param source: The source :class:`BaseContour` instance from which
            to copy data.

        Example::

            >>> contour.copyData(sourceContour)

        """
        super().copyData(source)
        for sourcePoint in source.points:
            self.appendPoint((0, 0))
            selfPoint = self.points[-1]
            selfPoint.copyData(sourcePoint)

    # -------
    # Parents
    # -------

    # Glyph

    _glyph: Callable[[], BaseGlyph] | None = None

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get or set the contour's parent glyph object.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the contour
            or :obj:`None`.
        :raises AssertionError: If attempting to set the glyph when it
            has already been set.

        Example::

            >>> glyph = contour.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph: BaseGlyph | Callable[[], BaseGlyph] | None) -> None:
        if self._glyph is not None:
            raise AssertionError("glyph for contour already set")
        if glyph is not None:
            glyph = reference(glyph)
        self._glyph = glyph

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the contour's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the contour
            or :obj:`None`.

        Example::

            >>> font = contour.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._glyph is None:
            return None
        return self.glyph.font

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the contour's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the contour
            or :obj:`None`.

        Example::

            >>> layer = contour.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._glyph is None:
            return None
        return self.glyph.layer

    # --------------
    # Identification
    # --------------

    # index

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get or set the index of the contour.

        The value must be an :class:`int`.

        :return: An :class:`int` representing the contour's index within an
            ordered list of the parent glyph's contours, or :obj:`None` if the
            contour does not belong to a glyph.
        :raises FontPartsError: If the contour does not belong to a glyph.

        Example::

            >>> contour.index
            1
            >>> contour.index = 0

        """,
    )

    def _get_base_index(self) -> int | None:
        glyph = self.glyph
        if glyph is None:
            return None
        value = self._get_index()
        value = normalizers.normalizeIndex(value)
        return value

    def _set_base_index(self, value: int) -> None:
        glyph = self.glyph
        if glyph is None:
            raise FontPartsError("The contour does not belong to a glyph.")
        normalizedValue = normalizers.normalizeIndex(value)
        if normalizedValue is None:
            return
        contourCount = len(glyph.contours)
        if normalizedValue < 0:
            normalizedValue = -(normalizedValue % contourCount)
        if normalizedValue >= contourCount:
            normalizedValue = contourCount
        self._set_index(normalizedValue)

    def _get_index(self) -> int | None:
        """Get the index of the native contour.

        This is the environment implementation of the :attr:`BaseContour.index`
        property getter.

        :return: An :class:`int` representing the contour's index within an
            ordered list of the parent glyph's contours, or :obj:`None` if the
            contour does not belong to a glyph. The value will be
            normalized with :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        glyph = self.glyph
        return glyph.contours.index(self)

    def _set_index(self, value: int) -> None:
        """Set the index of the contour.

        This is the environment implementation of the :attr:`BaseContour.index`
        property setter.

        :param value: The index to set as an :class:`int`. The value will have
            been normalized with :func:`normalizers.normalizeIndex`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # identifier

    def getIdentifierForPoint(self, point: BasePoint) -> str:
        """Generate and assign a unique identifier to the given point.

        If `point` already has an identifier, the existing identifier is returned.
        Otherwise, a new unique identifier is created and assigned to `point`.

        :param point: The :class:`BasePoint` instance to which the identifier
            should be assigned.
        :return: A :class:`str` representing the newly assigned identifier.

        Example::

            >>> contour.getIdentifierForPoint(point)
            'ILHGJlygfds'

        """
        point = normalizers.normalizePoint(point)
        return self._getIdentifierForPoint(point)

    def _getIdentifierForPoint(self, point: BasePoint) -> str:
        """Generate and assign a unique identifier to the given native point.

        This is the environment implementation
        of :meth:`BaseContour.getIdentifierForPoint`.

        :param point: The :class:`BasePoint` subclass instance to which the
            identifier should be assigned. The value will have been normalized
            with :func:`normalizers.normalizePoint`.
        :return: A :class:`str` representing the newly assigned identifier.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # ----
    # Pens
    # ----

    def draw(self, pen: PenType) -> None:
        """Draw the contour's outline data to the given pen.

        :param pen: The :class:`fontTools.pens.basePen.AbstractPen` to which the
            outline data should be drawn.

        Example::

            >>> contour.draw(pen)

        """
        self._draw(pen)

    def _draw(self, pen: PenType, **kwargs: Any) -> None:
        r"""Draw the native contour's outline data to the given pen.

        This is the environment implementation of :meth:`BaseContour.draw`.

        :param pen: The :class:`fontTools.pens.basePen.AbstractPen` to which the
            outline data should be drawn.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        from fontTools.ufoLib.pointPen import PointToSegmentPen

        adapter = PointToSegmentPen(pen)
        self.drawPoints(adapter)

    def drawPoints(self, pen: PointPenType) -> None:
        """Draw the contour's outline data to the given point pen.

        :param pen: The :class:`fontTools.pens.basePen.AbstractPointPen` to
            which the outline data should be drawn.

        Example::

            >>> contour.drawPoints(pointPen)

        """
        self._drawPoints(pen)

    def _drawPoints(self, pen: PointPenType, **kwargs: Any) -> None:
        r"""Draw the native contour's outline data to the given point pen.

        This is the environment implementation of :meth:`BaseContour.drawPoints`.

        :param pen: The :class:`fontTools.pens.basePen.AbstractPointPen` to
            which the outline data should be drawn.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        # The try: ... except TypeError: ...
        # handles backwards compatibility with
        # point pens that have not been upgraded
        # to point pen protocol 2.
        try:
            pen.beginPath(self.identifier)
        except TypeError:
            pen.beginPath()
        for point in self.points:
            typ = point.type
            if typ == "offcurve":
                typ = None
            try:
                pen.addPoint(
                    pt=(point.x, point.y),
                    segmentType=typ,
                    smooth=point.smooth,
                    name=point.name,
                    identifier=point.identifier,
                )
            except TypeError:
                pen.addPoint(
                    pt=(point.x, point.y),
                    segmentType=typ,
                    smooth=point.smooth,
                    name=point.name,
                )
        pen.endPath()

    # ------------------
    # Data normalization
    # ------------------

    def autoStartSegment(self) -> None:
        """Automatically calculate and set the contour's first segment.

        The behavior of this may vary across environments.

        Example::

            >>> contour.autoStartSegment()

        """
        self._autoStartSegment()

    def _autoStartSegment(self, **kwargs: Any) -> None:
        r"""Automatically calculate and set the native contour's first segment.

        This is the environment implementation of :meth:`BaseContour.autoStartSegment`.

        :param \**kwargs: Additional keyword arguments.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        startIndex = 0
        startSegment = self.segments[0]
        for i in range(len(self.segments)):
            segment = self.segments[i]
            startOn = startSegment.onCurve
            on = segment.onCurve
            if on.y <= startOn.y:
                if on.y == startOn.y:
                    if on.x < startOn.x:
                        startSegment = segment
                        startIndex = i
                else:
                    startSegment = segment
                    startIndex = i
        if startIndex != 0:
            self.setStartSegment(startIndex)

    def round(self) -> None:
        """Round all point coordinates in the contour to the nearest integer.

        Example::

            >>> contour.round()

        """
        self._round()

    def _round(self, **kwargs: Any) -> None:
        r"""Round all point coordinates in the native contour to the nearest integer.

        This is the environment implementation of :meth:`BaseContour.round`.

        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        for point in self.points:
            point.round()

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the contour according to the given matrix.

        This is the environment implementation of :meth:`BaseContour.transformBy`.

        :param matrix: The :ref:`type-transformation` to apply. The value will
             be normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        for point in self.points:
            point.transformBy(matrix)

    # -------------
    # Interpolation
    # -------------

    compatibilityReporterClass = ContourCompatibilityReporter

    def isCompatible(
        self, other: BaseContour, cls=None
    ) -> tuple[bool, ContourCompatibilityReporter]:
        """Evaluate interpolation compatibility with another contour.

        :param other: The other :class:`BaseContour` instance to check
            compatibility with.
        :return: A :class:`tuple` where the first element is a :class:`bool`
            indicating compatibility, and the second element is
            a :class:`fontParts.base.compatibility.ContourCompatibilityReporter`
            instance.

        Example::

            >>> compatible, report = self.isCompatible(otherContour)
            >>> compatible
            False
            >>> compatible
            [Fatal] Contour: [0] + [0]
            [Fatal] Contour: [0] contains 4 segments | [0] contains 3 segments
            [Fatal] Contour: [0] is closed | [0] is open

        """
        return super().isCompatible(other, BaseContour)

    def _isCompatible(
        self, other: BaseContour, reporter: ContourCompatibilityReporter
    ) -> None:
        """Evaluate interpolation compatibility with another native contour.

        This is the environment implementation of :meth:`BaseContour.isCompatible`.

        :param other: The other :class:`BaseContour` instance to check
            compatibility with.
        :param reporter: An object used to report compatibility issues.

        .. note::

            Subclasses may override this method.

        """
        contour1 = self
        contour2 = other
        # open/closed
        if contour1.open != contour2.open:
            reporter.openDifference = True
        # direction
        if contour1.clockwise != contour2.clockwise:
            reporter.directionDifference = True
        # segment count
        if len(contour1) != len(contour2.segments):
            reporter.segmentCountDifference = True
            reporter.fatal = True
        # segment pairs
        for i in range(min(len(contour1), len(contour2))):
            segment1 = contour1[i]
            segment2 = contour2[i]
            segmentCompatibility = segment1.isCompatible(segment2)[1]
            if segmentCompatibility.fatal or segmentCompatibility.warning:
                if segmentCompatibility.fatal:
                    reporter.fatal = True
                if segmentCompatibility.warning:
                    reporter.warning = True
                reporter.segments.append(segmentCompatibility)

    # ----
    # Open
    # ----

    open: dynamicProperty = dynamicProperty(
        "base_open",
        """Determine whether the contour is open.

        This property is read-only.

        :return: :obj:`True` if the contour is open, otherwise :obj:`False`.

        Example::

            >>> contour.open
            True

        """,
    )

    def _get_base_open(self) -> bool:
        value = self._get_open()
        value = normalizers.normalizeBoolean(value)
        return value

    def _get_open(self) -> bool:
        """Determine whether the native contour is open.

        This is the environment implementation of the :attr:`BaseContour.open`
        property getter.

        :return: :obj:`True` if the contour is open, otherwise :obj:`False`.
            The value will have been normalized
            with :func:`normalizers.normalizeBoolean`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # ---------
    # Direction
    # ---------

    clockwise: dynamicProperty = dynamicProperty(
        "base_clockwise",
        """Specify or determine whether the contour's winding direction is clockwise.

        The value must be a :class:`bool` indicating the contour's winding
        direction.

        :return: :obj:`True` if the contour's winding direction is clockwise,
            otherwise :obj:`False`.

        """,
    )

    def _get_base_clockwise(self) -> bool:
        value = self._get_clockwise()
        value = normalizers.normalizeBoolean(value)
        return value

    def _set_base_clockwise(self, value: bool) -> None:
        value = normalizers.normalizeBoolean(value)
        self._set_clockwise(value)

    def _get_clockwise(self) -> bool:
        """Determine whether the native contour's winding direction is clockwise.

        This is the environment implementation of the :attr:`BaseContour.clockwise`
        property getter.

        :return: :obj:`True` if the contour's winding direction is clockwise,
            otherwise :obj:`False`. The value will have been normalized with
            :func:`normalizers.normalizeBoolean`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_clockwise(self, value: bool) -> None:
        """Specify whether the native contour's winding direction is clockwise.

        This is the environment implementation of the :attr:`BaseContour.clockwise`
        property setter.

        :param value: A :class:`bool` indicating the desired winding
         direction. :obj:`True` sets the direction to clockwise,
         and :obj:`False` to counter-clockwise. The value will have been
         normalized with :func:`normalizers.normalizeBoolean`.

        .. note::

            Subclasses may override this method.

        """
        if self.clockwise != value:
            self.reverse()

    def reverse(self) -> None:
        """Reverse the direction of the contour.

        Example::

            >>> contour.clockwise
            False
            >>> contour.reverse()
            >>> contour.clockwise
            True

        """
        self._reverse()

    def _reverse(self, **kwargs) -> None:
        r"""Reverse the direction of the contour.

        This is the environment implementation of :meth:`BaseContour.reverse`.

        :param \**kwargs: Additional keyword arguments.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # ------------------------
    # Point and Contour Inside
    # ------------------------

    def pointInside(self, point: CoordinateLike) -> bool:
        """Check if `point` is within the filled area of the contour.

        :param point: The point to check as a :ref:`type-coordinate`.
        :return: :obj:`True` if `point` is inside the filled area of the
            contour, :obj:`False` otherwise.

        Example::

            >>> contour.pointInside((40, 65))
            True

        """
        point = normalizers.normalizeCoordinateTuple(point)
        return self._pointInside(point)

    def _pointInside(self, point: CoordinateLike) -> bool:
        """Check if `point` is within the filled area of the native contour.

         This is the environment implementation of :meth:`BaseContour.pointInside`.

        :param point: The point to check as a :ref:`type-coordinate`. The value
             will have been normalized with :func:`normalizers.normalizeCoordinateTuple`.
         :return: :obj:`True` if `point` is inside the filled area of the
             contour, :obj:`False` otherwise.

         .. note::

             Subclasses may override this method.

        """
        from fontTools.pens.pointInsidePen import PointInsidePen

        pen = PointInsidePen(glyphSet=None, testPoint=point, evenOdd=False)
        self.draw(pen)
        return pen.getResult()

    def contourInside(self, otherContour: BaseContour) -> bool:
        """Check if `otherContour` is within the current contour's filled area.

        :param point: The :class:`BaseContour` instance to check.
        :return: :obj:`True` if `otherContour` is inside the filled area of the
            current contour instance, :obj:`False` otherwise.

        Example::

            >>> contour.contourInside(otherContour)
            True

        """
        otherContour = normalizers.normalizeContour(otherContour)
        return self._contourInside(otherContour)

    def _contourInside(self, otherContour: BaseContour) -> bool:
        """Check if `otherContour` is within the current native contour's filled area.

        This is the environment implementation of :meth:`BaseContour.contourInside`.

        :param point: The :class:`BaseContour` instance to check. The value will have
            been normalized with :func:`normalizers.normalizeContour`.
        :return: :obj:`True` if `otherContour` is inside the filled area of the
            current contour instance, :obj:`False` otherwise.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # ---------------
    # Bounds and Area
    # ---------------

    bounds: dynamicProperty = dynamicProperty(
        "base_bounds",
        """Get the bounds of the contour.

        This property is read-only.

        :return: The :class:`Bounds` instance representing the the contour's bounds,
            or :obj:`None` if the component is empty.

        Example::

            >>> contour.bounds
            Bounds(xMin=10, yMin=30, xMax=765, yMax=643)


        """,
    )

    def _get_base_bounds(self) -> Bounds | None:
        value = self._get_bounds()
        if value is None:
            return None
        value = normalizers.normalizeBoundingBox(value)
        return Bounds(value)

    def _get_bounds(self) -> BoundingBox | None:
        """Get the bounds of the contour.

        This is the environment implementation of the :attr:`BaseContour.bounds`
        property getter.

        :return: A :class:`tuple` of four :class:`int` or :class:`float` values
            in the form ``(x minimum, y minimum, x maximum, y maximum)``
            representing the bounds of the contour, or :obj:`None` if the contour
            is open.

        .. note::

            Subclasses may override this method.

        """
        from fontTools.pens.boundsPen import BoundsPen

        pen = BoundsPen(self.layer)
        self.draw(pen)
        return pen.bounds

    area: dynamicProperty = dynamicProperty(
        "base_area",
        """Get the area of the contour

        This property is read-only.

        :return: A positive :class:`int` or a :class:` float value representing
            the area of the contour, or :obj:`None` if the contour is open.

        Example::

            >>> contour.area
            583

        """,
    )

    def _get_base_area(self) -> float | None:
        value = self._get_area()
        if value is not None:
            value = normalizers.normalizeArea(value)
        return value

    def _get_area(self) -> float | None:
        """Get the area of the native contour

        This is the environment implementation of the :attr:`BaseContour.area`
        property getter.

        :return: A positive :class:`int` or a :class:` float value representing
            the area of the contour, or :obj:`None` if the contour is open.

        .. note::

            Subclasses may override this method.

        """
        from fontTools.pens.areaPen import AreaPen

        pen = AreaPen(self.layer)
        self.draw(pen)
        if not pen.value:
            return None
        return abs(pen.value)

    # --------
    # Segments
    # --------

    # The base class implements the full segment interaction API.
    # Subclasses do not need to override anything within the contour
    # other than registering segmentClass. Subclasses may choose to
    # implement this API independently if desired.

    def _setContourInSegment(self, segment: BaseSegment) -> None:
        if segment.contour is None:
            segment.contour = self

    segments: dynamicProperty = dynamicProperty(
        "segments",
        """Get the contour's segments.

        This property is read-only.

        :return: A :class:`tuple` of :class:`BaseSegment` instances.

        Example::

            >>> contour.segments
            (<BaseSegment curve index='0' at 4573388368>, ...)

        """,
    )

    def _get_segments(self) -> tuple[BaseSegment, ...]:
        """Get the native contour's segments.

        This is the environment implementation of the :attr:`BaseContour.segments`
        property getter.

        :return: A :class:`tuple` of :class:`BaseSegment` subclass instances.

        .. note::

            Subclasses may override this method.

        """
        points = self.points
        if not points:
            return ()
        segments: list[list[BasePoint]] = [[]]
        lastWasOffCurve = False
        firstIsMove = points[0].type == "move"
        for point in points:
            segments[-1].append(point)
            if point.type != "offcurve":
                segments.append([])
            lastWasOffCurve = point.type == "offcurve"
        if len(segments[-1]) == 0:
            del segments[-1]
        if lastWasOffCurve and firstIsMove:
            # ignore trailing off curves
            del segments[-1]
        if lastWasOffCurve and not firstIsMove and len(segments) > 1:
            segment = segments.pop(-1)
            segment.extend(segments[0])
            del segments[0]
            segments.append(segment)
        if not lastWasOffCurve and not firstIsMove:
            segment = segments.pop(0)
            segments.append(segment)
        # wrap into segments
        wrapped: list[BaseSegment] = []
        for points in segments:
            if self.segmentClass is None:
                raise TypeError("segmentClass cannot be None.")
            pointSegment = self.segmentClass()
            pointSegment._setPoints(points)
            self._setContourInSegment(pointSegment)
            wrapped.append(pointSegment)
        return tuple(wrapped)

    def __getitem__(self, index: int) -> BaseSegment:
        """Get the segment at the specified index.

        :param index: The zero-based index of the point to retrieve as
            an :class:`int`.
        :return: The :class:`BaseSegment` instance located at the specified `index`.
        :raises IndexError: If the specified `index` is out of range.

        """
        return self.segments[index]

    def __iter__(self) -> Iterator[BaseSegment]:
        """Return an iterator over the segments in the contour.

        :return: An iterator over the :class:`BaseSegment` instances belonging
            to the contour.

        """
        return self._iterSegments()

    def _iterSegments(self) -> Iterator[BaseSegment]:
        segments = self.segments
        count = len(segments)
        index = 0
        while count:
            yield segments[index]
            count -= 1
            index += 1

    def __len__(self) -> int:
        """Return the number of segments in the contour.

        :return: An :class:`int` representing the number of :class:`BaseSegment`
            instances belonging to the contour.

        """
        return self._len__segments()

    def _len__segments(self, **kwargs: Any) -> int:
        r"""Return the number of segments in the native contour.

        This is the environment implementation of :meth:`BaseContour.__len__`.

        :return: An :class:`int` representing the number of :class:`BaseSegment`
            subclass instances belonging to the contour.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        return len(self.segments)

    def appendSegment(
        self,
        type: str | None = None,
        points: PointCollectionType | None = None,
        smooth: bool = False,
        segment: BaseSegment | None = None,
    ) -> None:
        """Append the given segment to the contour.

        If `type` or `points` are specified, those values will be used instead
        of the values in the given `segment` object. The specified `smooth`
        state will be applied if ``segment=None``.

        :param type: An optional :attr:`BaseSegment.type` to be applied to
            the segment as a :class:`str`. Defaults to :obj:`None`.
        :param points: The optional :attr:`BaseSegment.points` to be applied to
            the segment as a :class:`list` or :class:`tuple`
            of :ref:`type-coordinate` items. Defaults to :obj:`None`.
        :param smooth: The :attr:`BaseSegment.smooth` state to be applied to the
            segment as a :class:`bool`. Defaults to :obj:`False`.
        :param segment:  An optional :class:`BaseSegment` instance from which
            attribute values will be copied. Defualts to :obj:`None`.

        """
        if segment is not None:
            if type is None:
                type = segment.type
            if points is None:
                points = [(point.x, point.y) for point in segment.points]
            smooth = segment.smooth
        if type is None:
            raise TypeError("Type cannot be None.")
        type = normalizers.normalizeSegmentType(type)
        if points is None:
            raise TypeError("Points cannot be None.")
        normalizedPoints = [normalizers.normalizeCoordinateTuple(p) for p in points]
        # Avoid mypy invariant List error.
        castPoints = cast(PointCollectionType, normalizedPoints)
        smooth = normalizers.normalizeBoolean(smooth)
        self._appendSegment(type=type, points=castPoints, smooth=smooth)

    def _appendSegment(
        self, type: str, points: PointCollectionType, smooth: bool, **kwargs: Any
    ) -> None:
        r"""Append the given segment to the native contour.

        This is the environment implementation of :meth:`BaseContour.appendSegment`.

        :param type: The :attr:`BaseSegment.type` to be applied to the segment as
            a :class:`str`. The value will have been normalized
            with :func:`normalizers.normalizeSegmentType`.
        :param points: The :attr:`BaseSegment.points` to be applied to the segment as
            a :class:`list` or :class:`tuple` of :ref:`type-coordinate` items.
            The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :param smooth: The :attr:`BaseSegment.smooth` state to be applied to the segment
            as a :class:`bool`. The value will have been normalized
            with :func:`normalizers.normalizeBoolean`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        self._insertSegment(
            len(self) - 1, type=type, points=points, smooth=smooth, **kwargs
        )

    def insertSegment(
        self,
        index: int,
        type: str | None = None,
        points: PointCollectionType | None = None,
        smooth: bool = False,
        segment: BaseSegment | None = None,
    ) -> None:
        """Insert the given segment into the contour.

        If `type` or `points` are specified, those values will be used instead
        of the values in the given `segment` object. The specified `smooth`
        state will be applied if ``segment=None``.

        :param index: The :attr:`BaseSegment.index` to be applied to the segment
            as a :class:`int`.
        :param type: An optional :attr:`BaseSegment.type` to be applied to the
            segment as a :class:`str`. Defaults to :obj:`None`.
        :param points: The optional :attr:`BaseSegment.points` to be applied to
            the segment as a :class:`list` or :class:`tuple`
            of :ref:`type-coordinate` items. Defaults to :obj:`None`.
        :param smooth: The :attr:`BaseSegment.smooth` state to be applied to the
            segment as a :class:`bool`. Defaults to :obj:`False`.
        :param segment: An optional :class:`BaseSegment` instance from which
            attribute values will be copied. Defualts to :obj:`None`.

        """
        if segment is not None:
            if type is None:
                type = segment.type
            if points is None:
                points = [(point.x, point.y) for point in segment.points]
            smooth = segment.smooth
        normalizedIndex = normalizers.normalizeIndex(index)
        if normalizedIndex is None:
            raise TypeError("Index cannot be None.")
        if type is None:
            raise TypeError("Type cannot be None.")
        type = normalizers.normalizeSegmentType(type)
        if points is None:
            raise TypeError("Points cannot be None.")
        normalizedPoints = [normalizers.normalizeCoordinateTuple(p) for p in points]
        # Avoid mypy invariant List error.
        castPoints = cast(PointCollectionType, normalizedPoints)
        smooth = normalizers.normalizeBoolean(smooth)
        self._insertSegment(index=index, type=type, points=castPoints, smooth=smooth)

    def _insertSegment(
        self,
        index: int,
        type: str,
        points: PointCollectionType,
        smooth: bool,
        **kwargs: Any,
    ) -> None:
        r"""Insert the given segment into the native contour.

        This is the environment implementation of :meth:`BaseContour.insertSegment`.

        :param index: The :attr:`BaseSegment.index` to be applied to the segment
            as a :class:`int`. The value will have been normalized
            with :func:`normalizers.normalizeIndex`.
        :param type: The :attr:`BaseSegment.type` to be applied to the segment as
            a :class:`str`. The value will have been normalized
            with :func:`normalizers.normalizeSegmentType`.
        :param points: The :attr:`BaseSegment.points` to be applied to the segment as
            a :class:`list` or :class:`tuple` of :ref:`type-coordinate` items.
            The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :param smooth: The :attr:`BaseSegment.smooth` state to be applied to the
            segment as a :class:`bool`. The value will have been normalized
            with :func:`normalizers.normalizeBoolean`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        onCurve = points[-1]
        offCurve = points[:-1]
        segments = self.segments
        addPointCount = 1
        if self.open:
            index += 1
            addPointCount = 0
        ptCount = sum([len(segments[s].points) for s in range(index)]) + addPointCount
        self.insertPoint(ptCount, onCurve, type=type, smooth=smooth)
        for offCurvePoint in reversed(offCurve):
            self.insertPoint(ptCount, offCurvePoint, type="offcurve")

    def removeSegment(
        self, segment: BaseSegment | int, preserveCurve: bool = False
    ) -> None:
        """Remove the given segment from the contour.

        If ``preserveCurve=True``, an attempt will be made to preserve the
        overall shape of the curve after the segment is removed, provided the
        environment supports such functionality.

        :param segment: The segment to remove as a :class:`BaseSegment` instance,
            or an :class:`int` representing the segment's index.
        :param preserveCurve: A :class:`bool` indicating whether to preserve
            the curve's shape after the segment is removed. Defaults to :obj:`False`.
        :raises ValueError: If the segment index is out of range or if the
            specified segment is not part of the contour.

        Example::

            >>> contour.removeSegment(mySegment)
            >>> contour.removeSegment(2, preserveCurve=True)

        """
        if isinstance(segment, int):
            index = segment
        else:
            index = self.segments.index(segment)
        normalizedIndex = normalizers.normalizeIndex(index)
        if normalizedIndex is None:  # pragma: no cover
            return
        if normalizedIndex >= self._len__segments():
            raise ValueError(f"No segment located at index {normalizedIndex}.")
        preserveCurve = normalizers.normalizeBoolean(preserveCurve)
        self._removeSegment(normalizedIndex, preserveCurve)

    def _removeSegment(self, index: int, preserveCurve: bool, **kwargs: Any) -> None:
        r"""Remove the given segment from the native contour.

        This is the environment implementation of :meth:`BaseContour.removeSegment`.

        :param index: The segment to remove as an :class:`int` representing
            the segment's index. The value will have been normalized
            with :func:`normalizers.normalizeIndex`.
        :param preserveCurve: A :class:`bool` indicating whether to preserve
            the curve's shape after the segment is removed. Defaults to :obj:`False`.
            The value will have been normalized
            with :func:`normalizers.normalizeBoolean`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        segment = self.segments[index]
        for point in segment.points:
            self.removePoint(point, preserveCurve)

    def setStartSegment(self, segment: BaseSegment | int) -> None:
        """Set the first segment in the contour.

        :param segment: The segment to set as the first instance in the contour
            as a :class:`BaseSegment` instance, or an :class:`int` representing
            the segment's index.
        :raises FontPartsError: If the contour is open.
        :raises ValueError: If the segment index is out of range or if the
            specified segment is not part of the contour.

        Example::

            >>> contour.setStartSegment(mySegment)
            >>> contour.setStartSegment(2)

        """
        if self.open:
            raise FontPartsError("An open contour can not change the starting segment.")
        segments = self.segments
        if not isinstance(segment, int):
            segmentIndex = segments.index(segment)
        else:
            segmentIndex = segment
        if len(self.segments) < 2:
            return
        if segmentIndex == 0:
            return
        if segmentIndex >= len(segments):
            raise ValueError(
                f"The contour does not contain a segment at index {segmentIndex}"
            )
        self._setStartSegment(segmentIndex)

    def _setStartSegment(self, segmentIndex: int, **kwargs: Any) -> None:
        r"""Set the first segment in the native contour.

        This is the environment implementation of :meth:`BaseContour.setStartSegment`.

        :param segmentIndex: An :class:`int` representing the index of the
            segment to be set as the first instance in the contour.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        # get the previous segment and set
        # its on curve as the first point
        # in the contour. this matches the
        # iteration behavior of self.segments.
        segmentIndex -= 1
        segments = self.segments
        segment = segments[segmentIndex]
        self.setStartPoint(segment.points[-1])

    # -------
    # bPoints
    # -------

    bPoints: dynamicProperty = dynamicProperty(
        "bPoints",
        """Get a list of all bPoints in the contour.

        This property is read-only.

        :return: A :class:`tuple` of :class`BaseBPoints`.

        """,
    )

    def _get_bPoints(self) -> tuple[BaseBPoint, ...]:
        bPoints: list[BaseBPoint] = []
        for point in self.points:
            if point.type not in ("move", "line", "curve"):
                continue
            if self.bPointClass is None:
                raise TypeError("bPointClass cannot be None.")
            bPoint = self.bPointClass()
            bPoint.contour = self
            bPoint._setPoint(point)
            bPoints.append(bPoint)
        return tuple(bPoints)

    def appendBPoint(
        self,
        type: str | None = None,
        anchor: CoordinateLike | None = None,
        bcpIn: CoordinateLike | None = None,
        bcpOut: CoordinateLike | None = None,
        bPoint: BaseBPoint | None = None,
    ) -> None:
        """Append the given bPoint to the contour.

        If `type`, `anchor`, `bcpIn` or `bcpOut` are specified, those values
        will be used instead of the values in the given `bPoint` object.

        :param type: An optional :attr:`BaseBPoint.type` to be applied to
            the bPoint as a :class:`str`. Defaults to :obj:`None`.
        :param anchor: An optional :attr:`BaseBPoint.anchor` to be applied to
            the bPoint as a :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param bcpIn: An optional :attr:`BaseBPoint.bcpIn` to be applied to the
            bPoint as a :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param bcpOut: An optional :attr:`BaseBPoint.bcpOut` to be applied to the
            bPoint as a :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param bPoint: An optional :class:`BaseBPoint` instance from which
            attribute values will be copied. Defualts to :obj:`None`.

        """
        if bPoint is not None:
            if type is None:
                type = bPoint.type
            if anchor is None:
                anchor = bPoint.anchor
            if bcpIn is None:
                bcpIn = bPoint.bcpIn
            if bcpOut is None:
                bcpOut = bPoint.bcpOut
        if type is None:
            raise TypeError("Type cannot be None.")
        type = normalizers.normalizeBPointType(type)
        if anchor is None:
            raise TypeError("Anchor cannot be None.")
        anchor = normalizers.normalizeCoordinateTuple(anchor)
        if bcpIn is None:
            bcpIn = (0, 0)
        bcpIn = normalizers.normalizeCoordinateTuple(bcpIn)
        if bcpOut is None:
            bcpOut = (0, 0)
        bcpOut = normalizers.normalizeCoordinateTuple(bcpOut)
        self._appendBPoint(type, anchor, bcpIn=bcpIn, bcpOut=bcpOut)

    def _appendBPoint(
        self,
        type: str,
        anchor: CoordinateLike,
        bcpIn: CoordinateLike,
        bcpOut: CoordinateLike,
        **kwargs: Any,
    ) -> None:
        r"""Append the given bPoint to the native contour.

        This is the environment implementation of :meth:`BaseContour.appendBPoint`.

        :param type: The :attr:`BaseBPoint.type` to be applied to the bPoint as
            a :class:`str`. The value will have been normalized
            with :func:`normalizers.normalizeBPointType`.
        :param anchor: The :attr:`BaseBPoint.anchor` to be applied to the bPoint
            as a :ref:`type-coordinate`. The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :param bcpIn: The :attr:`BaseBPoint.bcpIn` to be applied to the bPoint
            as a :ref:`type-coordinate`. The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :param bcpOut: An optional :attr:`BaseBPoint.bcpOut` to be applied to
            the bPoint as a :ref:`type-coordinate`. The value will have been
            normalized with :func:`normalizers.normalizeCoordinateTuple`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        self.insertBPoint(len(self.bPoints), type, anchor, bcpIn=bcpIn, bcpOut=bcpOut)

    def insertBPoint(
        self,
        index: int,
        type: str | None = None,
        anchor: CoordinateLike | None = None,
        bcpIn: CoordinateLike | None = None,
        bcpOut: CoordinateLike | None = None,
        bPoint: BaseBPoint | None = None,
    ) -> None:
        """Insert the given bPoint into the contour.

        If `type`, `anchor`, `bcpIn` or `bcpOut` are specified, those values
        will be used instead of the values in the given `bPoint` object.

        :param index: The :attr:`BaseBPoint.index` to be applied to the bPoint
            as an :class:`int`.
        :param type: An optional :attr:`BaseBPoint.type` to be applied to
            the bPoint as a :class:`str`. Defaults to :obj:`None`.
        :param anchor: An optional :attr:`BaseBPoint.anchor` to be applied to
            the bPoint as a :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param bcpIn: An optional :attr:`BaseBPoint.bcpIn` to be applied to the
            bPoint as a :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param bcpOut: An optional :attr:`BaseBPoint.bcpOut` to be applied to the
            bPoint as a :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param bPoint: An optional :class:`BaseBPoint` instance from which
            attribute values will be copied. Defualts to :obj:`None`.

        """
        if bPoint is not None:
            if type is None:
                type = bPoint.type
            if anchor is None:
                anchor = bPoint.anchor
            if bcpIn is None:
                bcpIn = bPoint.bcpIn
            if bcpOut is None:
                bcpOut = bPoint.bcpOut
        normalizedIndex = normalizers.normalizeIndex(index)
        if normalizedIndex is None:
            raise TypeError("Index cannot be None.")
        if type is None:
            raise TypeError("Type cannot be None.")
        type = normalizers.normalizeBPointType(type)
        if anchor is None:
            raise TypeError("Anchor cannot be None.")
        anchor = normalizers.normalizeCoordinateTuple(anchor)
        if bcpIn is None:
            bcpIn = (0, 0)
        bcpIn = normalizers.normalizeCoordinateTuple(bcpIn)
        if bcpOut is None:
            bcpOut = (0, 0)
        bcpOut = normalizers.normalizeCoordinateTuple(bcpOut)
        self._insertBPoint(
            index=normalizedIndex, type=type, anchor=anchor, bcpIn=bcpIn, bcpOut=bcpOut
        )

    def _insertBPoint(
        self,
        index: int,
        type: str,
        anchor: CoordinateLike,
        bcpIn: CoordinateLike,
        bcpOut: CoordinateLike,
        **kwargs: Any,
    ) -> None:
        r"""Insert the given bPoint into the native contour.

        This is the environment implementation of :meth:`BaseContour.insertBPoint`.

        :param index: The :attr:`BaseBPoint.index` to be applied to the bPoint
            as an :class:`int`. The value will have been normalized
            with :func:`normalizers.normalizeIndex`.
        :param type: An optional :attr:`BaseBPoint.type` to be applied to
            the bPoint as a :class:`str`. The value will have been normalized
            with :func:`normalizers.normalizeBPointType`.
        :param anchor: The :attr:`BaseBPoint.anchor` to be applied to the bPoint
            as a :ref:`type-coordinate`. The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :param bcpIn: The :attr:`BaseBPoint.bcpIn` to be applied to the bPoint
            as a :ref:`type-coordinate`. The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :param bcpOut: An optional :attr:`BaseBPoint.bcpOut` to be applied to
            the bPoint as a :ref:`type-coordinate`. The value will have been
            normalized with :func:`normalizers.normalizeCoordinateTuple`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        # insert a simple line segment at the given anchor
        # look it up as a bPoint and change the bcpIn and bcpOut there
        # this avoids code duplication
        self._insertSegment(index=index, type="line", points=[anchor], smooth=False)
        bPoints = self.bPoints
        index += 1
        if index >= len(bPoints):
            # its an append instead of an insert
            # so take the last bPoint
            index = -1
        bPoint = bPoints[index]
        bPoint.type = type
        bPoint.bcpIn = bcpIn
        bPoint.bcpOut = bcpOut

    def removeBPoint(self, bPoint: BaseBPoint | int) -> None:
        """Remove the given bPoint from the contour.

        :param bPoint: The bPoint to remove as a :class:`BaseBPoint` instance,
            or an :class:`int` representing the bPoint's index.
        :raises ValueError: If the bPoint index is out of range or if the
            specified bPoint is not part of the contour.

        Example::

            >>> contour.removeBPoint(myBPoint)
            >>> contour.removeBPoint(2)

        """
        index = bPoint.index if not isinstance(bPoint, int) else bPoint
        normalizedIndex = normalizers.normalizeIndex(index)
        # Avoid mypy conflict with normalizeIndex -> Optional[int]
        if normalizedIndex is None:  # pragma: no cover
            return
        if normalizedIndex >= self._len__points():
            raise ValueError(f"No bPoint located at index {normalizedIndex}.")
        self._removeBPoint(normalizedIndex)

    def _removeBPoint(self, index: int, **kwargs: Any) -> None:
        r"""Remove the given bPoint from the native contour.

        This is the environment implementation of :meth:`BaseContour.removeBPoint`.

        :param index: The index representing the :class:`BaseBPoint` subclass
            instance to remove as an :class:`int`. The value will have been
            normalized with :func:`normalizers.normalizeIndex`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        bPoint = self.bPoints[index]

        nextSegment = bPoint._nextSegment
        offCurves = nextSegment.offCurve
        if offCurves:
            offCurve = offCurves[0]
            self.removePoint(offCurve)

        segment = bPoint._segment
        offCurves = segment.offCurve
        if offCurves:
            offCurve = offCurves[-1]
            self.removePoint(offCurve)

        self.removePoint(bPoint._point)

    # ------
    # Points
    # ------

    def _setContourInPoint(self, point: BasePoint) -> None:
        if point.contour is None:
            point.contour = self

    points: dynamicProperty = dynamicProperty(
        "points",
        """Get a list of all points in the contour.

        This property is read-only.

        :return: A :class:`tuple` of :class`BasePoints`.

        """,
    )

    def _get_points(self) -> tuple[BasePoint, ...]:
        """Get a list of all points in the native contour.

        This is the environment implementation of the :attr:`BaseContour.points`
        property getter.

        :return: A :class:`tuple` of :class`BasePoint` subclass instances.

        .. note::

            Subclasses may override this method.

        """
        return tuple(self._getitem__points(i) for i in range(self._len__points()))

    def _len__points(self) -> int:
        return self._lenPoints()

    def _lenPoints(self, **kwargs: Any) -> int:
        r"""Return the number of points in the native contour.

        :param \**kwargs: Additional keyword arguments.
        :return: An :class:`int` representing the number of :class:`BasePoint`
            subclass instances belonging to the contour.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _getitem__points(self, index: int) -> BasePoint:
        normalizedIndex = normalizers.normalizeIndex(index)
        if normalizedIndex is None or normalizedIndex >= self._len__points():
            raise ValueError(f"No point located at index {normalizedIndex}.")
        point = self._getPoint(normalizedIndex)
        self._setContourInPoint(point)
        return point

    def _getPoint(self, index: int, **kwargs: Any) -> BasePoint:
        r"""Get the given point from the native contour.

        :param index: The index representing the :class:`BaseBPoint` subclass
            instance to retrieve as an :class:`int`. The value will have been
            normalized with :func:`normalizers.normalizeIndex`.
        :param \**kwargs: Additional keyword arguments.
        :return: A :class:`BasePoint` subclass instance.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _getPointIndex(self, point: BasePoint) -> int:
        for i, other in enumerate(self.points):
            if point == other:
                return i
        raise FontPartsError("The point could not be found.")

    def appendPoint(
        self,
        position: CoordinateLike | None = None,
        type: str | None = None,
        smooth: bool = False,
        name: str | None = None,
        identifier: str | None = None,
        point: BasePoint | None = None,
    ) -> None:
        """Append the given point to the contour.

        If `position`, `type`, `name` or identifier are specified, those values will be
        used instead of the values in the given `point` object.

        If `type` is still not resolved (i.e., ``type=None`` and no `point` is given),
        it defaults to ``"line"``.

        The specified `smooth` state will be applied if ``point=None``.

        :param position: An optional position to be applied to the point as a
            :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param type: An optional :attr:`BasePoint.type` to be applied to the point as a
            :class:`str`. Defaults to :obj:`None`.
        :param smooth: The :attr:`BasePoint.smooth` state to be applied to the point as
            a :class:`bool`. Defaults to :obj:`False`.
        :param name: An optional :attr:`BasePoint.name` to be applied to the point as a
            :class:`str`. Defaults to :obj:`None`.
        :param identifier: An optional :attr:`BasePoint.identifier` to be applied to the
            point as a :class:`str`. Defaults to :obj:`None`.
        :param point: An optional :class:`BasePoint` instance from which attribute
            values will be copied. Defualts to :obj:`None`.

        """
        if point is not None:
            if position is None:
                position = point.position
            if type is None:
                type = point.type if point.type is not None else "line"
            smooth = point.smooth
            if name is None:
                name = point.name
            if identifier is None:
                identifier = point.identifier
        self.insertPoint(
            len(self.points),
            position=position,
            type=type,
            smooth=smooth,
            name=name,
            identifier=identifier,
        )

    def insertPoint(
        self,
        index: int,
        position: CoordinateLike | None = None,
        type: str | None = None,
        smooth: bool = False,
        name: str | None = None,
        identifier: str | None = None,
        point: BasePoint | None = None,
    ) -> None:
        """Insert the given point into the contour.

        If `position`, `type`, `name` or `identifier` are specified, those values will
        be used instead of the values in the given `point` object.

        If `type` is still not resolved (i.e., ``type=None`` and no `point` is given),
        it defaults to ``"line"``.

        The specified `smooth` state will be applied if ``point=None``.

        :param index: The :attr:`BasePoint.index` to be applied to the point as an
            :class:`int`.
        :param position: An optional position to be applied to the point as a
            :ref:`type-coordinate`. Defaults to :obj:`None`.
        :param type: An optional :attr:`BasePoint.type` to be applied to the point as a
            :class:`str`. Defaults to :obj:`None`.
        :param smooth: The :attr:`BasePoint.smooth` state to be applied to the point as
            a :class:`bool`. Defaults to :obj:`False`.
        :param name: An optional :attr:`BasePoint.name` to be applied to the point as a
            :class:`str`. Defaults to :obj:`None`.
        :param identifier: An optional :attr:`BasePoint.identifier` to be applied to the
            point as a :class:`str`. Defaults to :obj:`None`.
        :param point: An optional :class:`BasePoint` instance from which attribute
            values will be copied. Defaults to :obj:`None`.

        """
        if point is not None:
            if position is None:
                position = point.position
            if type is None:
                type = point.type
            smooth = point.smooth
            if name is None:
                name = point.name
            if identifier is None:
                identifier = point.identifier
        normalizedIndex = normalizers.normalizeIndex(index)
        if normalizedIndex is None:
            raise TypeError("Index cannot be None.")
        if position is None:
            raise TypeError("Position cannot be None.")
        position = normalizers.normalizeCoordinateTuple(position)
        type = "line" if type is None else type
        type = normalizers.normalizePointType(type)
        smooth = normalizers.normalizeBoolean(smooth)
        if name is not None:
            name = normalizers.normalizePointName(name)
        if identifier is not None:
            identifier = normalizers.normalizeIdentifier(identifier)
        self._insertPoint(
            normalizedIndex,
            position=position,
            type=type,
            smooth=smooth,
            name=name,
            identifier=identifier,
        )

    def _insertPoint(
        self,
        index: int,
        position: CoordinateLike,
        type: str,
        smooth: bool,
        name: str | None,
        identifier: str | None,
        **kwargs: Any,
    ) -> None:
        r"""Insert the given point into the native contour.

        This is the environment implementation of :meth:`BaseContour.insertPoint`.

        :param index: The :attr:`BasePoint.index` to be applied to the point
            as an :class:`int`. The value will have been normalized
            with :func:`normalizers.normalizeIndex`.
        :param position: The position to be applied to the point as
            a :ref:`type-coordinate`. The value will have been normalized with
            :func:`normalizers.normalizeCoordinateTuple`.
        :param type: The :attr:`BasePoint.type` to be applied to the point as
            a :class:`str`. The value will have been normalized
            with :func:`normalizers.normalizePointType`.
        :param smooth: The :attr:`BasePoint.smooth` state to be applied to the
            point as a :class:`bool`. The value will have been normalized
            with :func:`normalizers.normalizeBoolean`.
        :param name: An optional :attr:`BasePoint.name` to be applied to the
            point as a :class:`str`. The value will have been normalized
            with :func:`normalizers.normalizePointName`
        :param identifier: An optional :attr:`BasePoint.identifier` to be
            applied to the point as a :class:`str`. The value will have been
            normalized with :func:`normalizers.normalizeIdentifier`, but will
            not have been tested for uniqueness.
        :param \**kwargs: Additional keyword arguments.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def removePoint(self, point: BasePoint | int, preserveCurve: bool = False) -> None:
        """Remove the given point from the contour.

        If ``preserveCurve=True``, an attempt will be made to preserve the
        overall shape of the curve after the segment is removed, provided the
        environment supports such functionality.

        :param point: The point to remove as a :class:`BasePoint` instance,
            or an :class:`int` representing the point's index.
        :param preserveCurve: A :class:`bool` indicating whether to preserve
            the curve's shape after the point is removed. Defaults to :obj:`False`.
        :raises ValueError: If the point index is out of range or if the
            specified point is not part of the contour.

        Example::

            >>> contour.removePoint(myPoint)
            >>> contour.removePoint(2, preserveCurve=True)

        """
        index = self.points.index(point) if not isinstance(point, int) else point
        normalizedIndex = normalizers.normalizeIndex(index)
        # Avoid mypy conflict with normalizeIndex -> Optional[int]
        if normalizedIndex is None:  # pragma: no cover
            return
        if normalizedIndex >= self._len__points():
            raise ValueError(f"No point located at index {normalizedIndex}.")
        preserveCurve = normalizers.normalizeBoolean(preserveCurve)
        self._removePoint(normalizedIndex, preserveCurve)

    def _removePoint(self, index: int, preserveCurve: bool, **kwargs: Any) -> None:
        r"""Remove the given point from the native contour.

        This is the environment implementation of :meth:`BaseContour.removePoint`.

        :param index: The index representing the :class:`BasePoint` subclass
            instance to remove as an :class:`int`. The value will have been
            normalized with :func:`normalizers.normalizeIndex`.
        :param preserveCurve: A :class:`bool` indicating whether to preserve
            the curve's shape after the point is removed. The value will have been
            normalized with :func:`normalizers.normalizeBoolean`.
        :param \**kwargs: Additional keyword arguments.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def setStartPoint(self, point: BasePoint | int) -> None:
        """Set the first segment in the contour.

        :param segment: The point to set as the first instance in the contour
            as a :class:`BasePoint` instance, or an :class:`int` representing
            the point's index.
        :raises FontPartsError: If the contour is open.
        :raises ValueError: If the point index is out of range or if the
            specified point is not part of the contour.

        Example::

            >>> contour.setStartPoint(myPoint)
            >>> contour.setStartPoint(2)

        """
        if self.open:
            raise FontPartsError("An open contour can not change the starting point.")
        points = self.points
        if not isinstance(point, int):
            pointIndex = points.index(point)
        else:
            pointIndex = point
        if pointIndex == 0:
            return
        if pointIndex >= len(points):
            raise ValueError(
                f"The contour does not contain a point at index {pointIndex}"
            )
        self._setStartPoint(pointIndex)

    def _setStartPoint(self, pointIndex: int, **kwargs: Any) -> None:
        r"""Set the first segment in the native contour.

        This is the environment implementation of :meth:`BaseContour.setStartPoint`.

        :param pointIndex: An :class:`int` representing the index of the point
            to be set as the first instance in the contour.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        points = self.points
        points = points[pointIndex:] + points[:pointIndex]
        # Clear the points.
        for point in self.points:
            self.removePoint(point)
        # Add the points.
        for point in points:
            self.appendPoint(
                (point.x, point.y),
                type=point.type,
                smooth=point.smooth,
                name=point.name,
                identifier=point.identifier,
            )

    # ---------
    # Selection
    # ---------

    # segments

    selectedSegments: dynamicProperty = dynamicProperty(
        "base_selectedSegments",
        """Get or set the selected segments in the contour.

        The value must be a :class:`tuple` or :class:`list` of
        either :class:`BaseSegment` instances or :class:`int` values
        representing segment indexes to select.

        :return: A :class:`tuple` of the currently selected :class:`BaseSegment`
            instances.

        Getting selected segments::

            >>> for segment in contour.selectedSegments:
            ...     segment.move((10, 20))

        Setting selected segments::

            >>> contour.selectedSegments = someSegments

        Setting selection using indexes::

            >>> contour.selectedSegments = [0, 2]

        """,
    )

    def _get_base_selectedSegments(self) -> tuple[BaseSegment, ...]:
        selected = tuple(
            normalizers.normalizeSegment(segment)
            for segment in self._get_selectedSegments()
        )
        return selected

    def _get_selectedSegments(self) -> tuple[BaseSegment, ...]:
        """Get the selected segments in the native contour.

        This is the environment implementation of the
        :attr:`BaseContour.selectedSegments` property getter.

        :return: A :class:`tuple` of the currently selected :class:`BaseSegment`
            instances. Each value item will be normalized
            with :func:`normalizers.normalizeSegment`.

        .. note::

            Subclasses may override this method.

        """
        return self._getSelectedSubObjects(self.segments)

    def _set_base_selectedSegments(
        self, value: CollectionType[BaseSegment | int]
    ) -> None:
        normalized = []
        for segment in value:
            normalizedSegment: BaseSegment | int
            if isinstance(segment, int):
                normalizedIndex = normalizers.normalizeIndex(segment)
                # Avoid mypy conflict with normalizeIndex -> Optional[int]
                if normalizedIndex is None:  # pragma: no cover
                    continue
                normalizedSegment = normalizedIndex
            else:
                normalizedSegment = normalizers.normalizeSegment(segment)
            normalized.append(normalizedSegment)
        self._set_selectedSegments(normalized)

    def _set_selectedSegments(self, value: CollectionType[BaseSegment | int]) -> None:
        """Set the selected segments in the native contour.

        This is the environment implementation of the
        :attr:`BaseContour.selectedSegments` property setter.

        :param value: The segments to select as a :class:`tuple`
            or :class:`list` of either :class:`BaseContour` instances
            or :class:`int` values representing segment indexes. Each value item
            will have been normalized with :func:`normalizers.normalizeSegment`
            or :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        return self._setSelectedSubObjects(self.segments, value)

    # points

    selectedPoints: dynamicProperty = dynamicProperty(
        "base_selectedPoints",
        """Get or set the selected points in the contour.

        The value must be a :class:`tuple` or :class:`list` of
        either :class:`BasePoint` instances or :class:`int` values
        representing point indexes to select.

        :return: A :class:`tuple` of the currently selected :class:`BasePoint`
            instances.

        Getting selected points::

            >>> for point in contour.selectedPoints:
            ...     point.move((10, 20))

        Setting selected points::

            >>> contour.selectedPoints = somePoints

        Setting selection using indexes::

            >>> contour.selectedPoints = [0, 2]

        """,
    )

    def _get_base_selectedPoints(self) -> tuple[BasePoint, ...]:
        selected = tuple(
            normalizers.normalizePoint(point) for point in self._get_selectedPoints()
        )
        return selected

    def _get_selectedPoints(self) -> tuple[BasePoint, ...]:
        """Get the selected points in the native contour.

        This is the environment implementation of
        the :attr:`BaseContour.selectedPoints` property getter.

        :return: A :class:`tuple` of the currently selected :class:`BasePoint`
            instances. Each value item will be normalized
            with :func:`normalizers.normalizePoint`.

        .. note::

            Subclasses may override this method.

        """
        return self._getSelectedSubObjects(self.points)

    def _set_base_selectedPoints(self, value: CollectionType[BasePoint | int]) -> None:
        normalized = []
        for point in value:
            normalizedPoint: BasePoint | int
            if isinstance(point, int):
                normalizedIndex = normalizers.normalizeIndex(point)
                # Avoid mypy conflict with normalizeIndex -> Optional[int]
                if normalizedIndex is None:  # pragma: no cover
                    continue
                normalizedPoint = normalizedIndex
            else:
                normalizedPoint = normalizers.normalizePoint(point)
            normalized.append(normalizedPoint)
        self._set_selectedPoints(normalized)

    def _set_selectedPoints(self, value: CollectionType[BasePoint | int]) -> None:
        """Set the selected points in the native contour.

        This is the environment implementation of
        the :attr:`BaseContour.selectedPoints` property setter.

        :param value: The points to select as a :class:`tuple` or :class:`list`
            of either :class:`BasePoint` instances or :class:`int` values
            representing point indexes to select. Each value item will have been
            normalized with :func:`normalizers.normalizePoint`
            or :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        return self._setSelectedSubObjects(self.points, value)

    # bPoints

    selectedBPoints: dynamicProperty = dynamicProperty(
        "base_selectedBPoints",
        """Get or set the selected bPoints in the contour.

        The value must be a :class:`tuple` or :class:`list` of
        either :class:`BaseBPoint` instances or :class:`int` values
        representing bPoint indexes to select.

        :return: A :class:`tuple` of the currently selected :class:`BaseBPoint`
            instances.

        Getting selected bPoints::

            >>> for bPoint in contour.selectedBPoints:
            ...     bPoint.move((10, 20))

        Setting selected bPoints::

            >>> contour.selectedBPoints = someBPoints

        Setting selection using indexes::

            >>> contour.selectedBPoints = [0, 2]

        """,
    )

    def _get_base_selectedBPoints(self) -> tuple[BaseBPoint, ...]:
        selected = tuple(
            normalizers.normalizeBPoint(bPoint)
            for bPoint in self._get_selectedBPoints()
        )
        return selected

    def _get_selectedBPoints(self) -> tuple[BaseBPoint, ...]:
        """Get the selected bPoints in the native contour.

        This is the environment implementation of
        the :attr:`BaseContour.selectedBPoints` property getter.

        :return: A :class:`tuple` of the currently selected :class:`BaseBPoint`
            instances. Each value item will be normalized
            with :func:`normalizers.normalizeBPoint`.

        .. note::

            Subclasses may override this method.

        """
        return self._getSelectedSubObjects(self.bPoints)

    def _set_base_selectedBPoints(
        self, value: CollectionType[BaseBPoint | int]
    ) -> None:
        normalized = []
        for bPoint in value:
            normalizedBPoint: BaseBPoint | int
            if isinstance(bPoint, int):
                normalizedIndex = normalizers.normalizeIndex(bPoint)
                # Avoid mypy conflict with normalizeIndex -> Optional[int]
                if normalizedIndex is None:  # pragma: no cover
                    continue
                normalizedBPoint = normalizedIndex
            else:
                normalizedBPoint = normalizers.normalizeBPoint(bPoint)
            normalized.append(normalizedBPoint)
        self._set_selectedBPoints(normalized)

    def _set_selectedBPoints(self, value: CollectionType[BaseBPoint | int]) -> None:
        """Set the selected bPoints in the native contour.

        This is the environment implementation of
        the :attr:`BaseContour.selectedBPoints` property setter.

        :param value: The bPoints to select as a :class:`tuple` or :class:`list`
            of either :class:`BaseBPoint` instances or :class:`int` values
            representing bPoint indexes to select. Each value item will have been
            normalized with :func:`normalizers.normalizeBPoint`
            or :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        return self._setSelectedSubObjects(self.bPoints, value)
